"""
EvAdventure Twitch-based combat

This implements a 'twitch' (aka DIKU or other traditional muds) style of MUD combat.

----

"""
from evennia import AttributeProperty, CmdSet, default_cmds
from evennia.commands.command import Command, InterruptCommand
from evennia.utils.utils import (
    display_len,
    inherits_from,
    list_to_string,
    pad,
    repeat,
    unrepeat,
)

from .characters import EvAdventureCharacter
from .combat_base import (
    CombatActionAttack,
    CombatActionHold,
    CombatActionStunt,
    CombatActionUseItem,
    CombatActionWield,
    EvAdventureCombatBaseHandler,
)
from .enums import ABILITY_REVERSE_MAP


class EvAdventureCombatTwitchHandler(EvAdventureCombatBaseHandler):
    """
    This is created on the combatant when combat starts. It tracks only the combatants
    side of the combat and handles when the next action will happen.


    """

    # fixed properties
    action_classes = {
        "hold": CombatActionHold,
        "attack": CombatActionAttack,
        "stunt": CombatActionStunt,
        "use": CombatActionUseItem,
        "wield": CombatActionWield,
    }

    # dynamic properties

    advantage_against = AttributeProperty(dict)
    disadvantage_against = AttributeProperty(dict)

    action_dict = AttributeProperty(dict)
    fallback_action_dict = AttributeProperty({"key": "hold", "dt": 0})

    # stores the current ticker reference, so we can manipulate it later
    current_ticker_ref = AttributeProperty(None)

    def msg(self, message, broadcast=True, **kwargs):
        """
        Central place for sending messages to combatants. This allows
        for adding any combat-specific text-decoration in one place.

        Args:
            message (str): The message to send.
            combatant (Object): The 'You' in the message, if any.
            broadcast (bool): If `False`, `combatant` must be included and
                will be the only one to see the message. If `True`, send to
                everyone in the location.
            location (Object, optional): If given, use this as the location to
                send broadcast messages to. If not, use `self.obj` as that
                location.

        Notes:
            If `combatant` is given, use `$You/you()` markup to create
            a message that looks different depending on who sees it. Use
            `$You(combatant_key)` to refer to other combatants.
        """
        super().msg(message, combatant=self.obj, broadcast=broadcast, location=self.obj.location)

    def at_init(self):
        self.obj.cmdset.add(TwitchLookCmdSet, persistent=False)

    def get_sides(self, combatant):
        """
        Get a listing of the two 'sides' of this combat, from the perspective of the provided
        combatant. The sides don't need to be balanced.

        Args:
            combatant (Character or NPC): The one whose sides are to determined.

        Returns:
            tuple: A tuple of lists `(allies, enemies)`, from the perspective of `combatant`.
                Note that combatant itself is not included in either of these.

        """
        # get all entities involved in combat by looking up their combathandlers
        combatants = [
            comb
            for comb in self.obj.location.contents
            if hasattr(comb, "scripts") and comb.scripts.has(self.key)
        ]
        location = self.obj.location

        if hasattr(location, "allow_pvp") and location.allow_pvp:
            # in pvp, everyone else is an enemy
            allies = [combatant]
            enemies = [comb for comb in combatants if comb != combatant]
        else:
            # otherwise, enemies/allies depend on who combatant is
            pcs = [comb for comb in combatants if inherits_from(comb, EvAdventureCharacter)]
            npcs = [comb for comb in combatants if comb not in pcs]
            if combatant in pcs:
                # combatant is a PC, so NPCs are all enemies
                allies = pcs
                enemies = npcs
            else:
                # combatant is an NPC, so PCs are all enemies
                allies = npcs
                enemies = pcs
        return allies, enemies

    def give_advantage(self, recipient, target):
        """
        Let a benefiter gain advantage against the target.

        Args:
            recipient (Character or NPC): The one to gain the advantage. This may or may not
                be the same entity that creates the advantage in the first place.
            target (Character or NPC): The one against which the target gains advantage. This
                could (in principle) be the same as the benefiter (e.g. gaining advantage on
                some future boost)

        """
        self.advantage_against[target] = True

    def give_disadvantage(self, recipient, target):
        """
        Let an affected party gain disadvantage against a target.

        Args:
            recipient (Character or NPC): The one to get the disadvantage.
            target (Character or NPC): The one against which the target gains disadvantage, usually
                an enemy.

        """
        self.disadvantage_against[target] = True

    def has_advantage(self, combatant, target):
        """
        Check if a given combatant has advantage against a target.

        Args:
            combatant (Character or NPC): The one to check if they have advantage
            target (Character or NPC): The target to check advantage against.

        """
        return self.advantage_against.get(target, False)

    def has_disadvantage(self, combatant, target):
        """
        Check if a given combatant has disadvantage against a target.

        Args:
            combatant (Character or NPC): The one to check if they have disadvantage
            target (Character or NPC): The target to check disadvantage against.

        """
        return self.disadvantage_against.get(target, False)

    def queue_action(self, action_dict, combatant=None):
        """
        Schedule the next action to fire.

        Args:
            action_dict (dict): The new action-dict to initialize.
            combatant: Unused.

        """
        if action_dict["key"] not in self.action_classes:
            self.obj.msg("This is an unkown action!")
            return

        # store action dict and schedule it to run in dt time
        self.action_dict = action_dict
        dt = action_dict.get("dt", 0)

        if self.current_ticker_ref:
            # we already have a current ticker going - abort it
            unrepeat(self.current_ticker_ref)
        if dt <= 0:
            # no repeat
            self.current_ticker_ref = None
        else:
            # always schedule the task to be repeating, cancel later otherwise. We store
            # the tickerhandler's ref to make sure we can remove it later
            self.current_ticker_ref = repeat(dt, self.execute_next_action, id_string="combat")

    def execute_next_action(self):
        """
        Triggered after a delay by the command
        """
        combatant = self.obj
        action_dict = self.action_dict
        action_class = self.action_classes[action_dict["key"]]
        action = action_class(self, combatant, action_dict)

        if action.can_use():
            action.execute()
            action.post_execute()

        if not action_dict.get("repeat", True):
            # not a repeating action, use the fallback (normally the original attack)
            self.action_dict = self.fallback_action_dict
            self.queue_action(self.fallback_action_dict)

        self.check_stop_combat()

    def check_stop_combat(self):
        """
        Check if the combat is over.
        """

        allies, enemies = self.get_sides(self.obj)

        location = self.obj.location

        # only keep combatants that are alive and still in the same room
        allies = [comb for comb in allies if comb.hp > 0 and comb.location == location]
        enemies = [comb for comb in enemies if comb.hp > 0 and comb.location == location]

        if not allies and not enemies:
            self.msg("Noone stands after the dust settles.", broadcast=False)
            self.stop_combat()
            return

        if not allies or not enemies:
            if allies + enemies == [self.obj]:
                self.msg("The combat is over.")
            else:
                still_standing = list_to_string(f"$You({comb.key})" for comb in allies + enemies)
                self.msg(
                    f"The combat is over. Still standing: {still_standing}.",
                    broadcast=False,
                )
            self.stop_combat()

    def stop_combat(self):
        """
        Stop combat immediately.
        """
        self.queue_action({"key": "hold", "dt": 0})  # make sure ticker is killed
        del self.obj.ndb.combathandler
        self.obj.cmdset.remove(TwitchLookCmdSet)
        self.delete()


class _BaseTwitchCombatCommand(Command):
    """
    Parent class for all twitch-combat commnads.

    """

    def at_pre_command(self):
        """
        Called before parsing.

        """
        if not self.caller.location or not self.caller.location.allow_combat:
            self.msg("Can't fight here!")
            raise InterruptCommand()

    def parse(self):
        """
        Handle parsing of most supported combat syntaxes (except stunts).

        <action> [<target>|<item>]
        or
        <action> <item> [on] <target>

        Use 'on' to differentiate if names/items have spaces in the name.

        """
        self.args = args = self.args.strip()
        self.lhs, self.rhs = "", ""

        if not args:
            return

        if " on " in args:
            lhs, rhs = args.split(" on ", 1)
        else:
            lhs, *rhs = args.split(None, 1)
            rhs = " ".join(rhs)
        self.lhs, self.rhs = lhs.strip(), rhs.strip()

    def get_or_create_combathandler(self, target=None, combathandler_key="combathandler"):
        """
        Get or create the combathandler assigned to this combatant.

        """
        if target:
            # add/check combathandler to the target
            if target.hp_max is None:
                self.msg("You can't attack that!")
                raise InterruptCommand()

            EvAdventureCombatTwitchHandler.get_or_create_combathandler(
                target, key=combathandler_key
            )
        return EvAdventureCombatTwitchHandler.get_or_create_combathandler(self.caller)


class CmdAttack(_BaseTwitchCombatCommand):
    """
    Attack a target. Will keep attacking the target until
    combat ends or another combat action is taken.

    Usage:
        attack/hit <target>

    """

    key = "attack"
    aliases = ["hit"]
    help_category = "combat"

    def func(self):
        target = self.caller.search(self.lhs)
        if not target:
            return

        combathandler = self.get_or_create_combathandler(target)
        # we use a fixed dt of 3 here, to mimic Diku style; one could also picture
        # attacking at a different rate, depending on skills/weapon etc.
        combathandler.queue_action({"key": "attack", "target": target, "dt": 3, "repeat": True})
        combathandler.msg(f"$You() $conj(attack) $You({target.key})!", self.caller)


class CmdLook(default_cmds.CmdLook, _BaseTwitchCombatCommand):
    def func(self):
        # get regular look, followed by a combat summary
        super().func()
        if not self.args:
            combathandler = self.get_or_create_combathandler()
            txt = str(combathandler.get_combat_summary(self.caller))
            maxwidth = max(display_len(line) for line in txt.strip().split("\n"))
            self.msg(f"|r{pad(' Combat Status ', width=maxwidth, fillchar='-')}|n\n{txt}")


class CmdHold(_BaseTwitchCombatCommand):
    """
    Hold back your blows, doing nothing.

    Usage:
        hold

    """

    key = "hold"

    def func(self):
        combathandler = self.get_or_create_combathandler()
        combathandler.queue_action({"key": "hold"})
        combathandler.msg("$You() $conj(hold) back, doing nothing.", self.caller)


class CmdStunt(_BaseTwitchCombatCommand):
    """
    Perform a combat stunt, that boosts an ally against a target, or
    foils an enemy, giving them disadvantage against an ally.

    Usage:
        boost [ability] <recipient> <target>
        foil [ability] <recipient> <target>
        boost [ability] <target>       (same as boost me <target>)
        foil [ability] <target>        (same as foil <target> me)

    Example:
        boost STR me Goblin
        boost DEX Goblin
        foil STR Goblin me
        foil INT Goblin
        boost INT Wizard Goblin

    """

    key = "stunt"
    aliases = (
        "boost",
        "foil",
    )
    help_category = "combat"

    def parse(self):
        args = self.args

        if not args or " " not in args:
            self.msg("Usage: <ability> <recipient> <target>")
            raise InterruptCommand()

        advantage = self.cmdname != "foil"

        # extract data from the input

        stunt_type, recipient, target = None, None, None

        stunt_type, *args = args.split(None, 1)
        if stunt_type:
            stunt_type = stunt_type.strip().lower()

        args = args[0] if args else ""

        recipient, *args = args.split(None, 1)
        target = args[0] if args else None

        # validate input and try to guess if not given

        # ability is requried
        if not stunt_type or stunt_type not in ABILITY_REVERSE_MAP:
            self.msg(
                f"'{stunt_type}' is not a valid ability. Pick one of"
                f" {', '.join(ABILITY_REVERSE_MAP.keys())}."
            )
            raise InterruptCommand()

        if not recipient:
            self.msg("Must give at least a recipient or target.")
            raise InterruptCommand()

        if not target:
            # something like `boost str target`
            target = recipient if advantage else "me"
            recipient = "me" if advantage else recipient

        # if we still have None:s at this point, we can't continue
        if None in (stunt_type, recipient, target):
            self.msg("Both ability, recipient and  target of stunt must be given.")
            raise InterruptCommand()

        # save what we found so it can be accessed from func()
        self.advantage = advantage
        self.stunt_type = ABILITY_REVERSE_MAP[stunt_type]
        self.recipient = recipient.strip()
        self.target = target.strip()

    def func(self):
        target = self.caller.search(self.target)
        if not target:
            return
        recipient = self.caller.search(self.recipient)
        if not recipient:
            return

        combathandler = self.get_or_create_combathandler(target)

        combathandler.queue_action(
            {
                "key": "stunt",
                "recipient": recipient,
                "target": target,
                "advantage": self.advantage,
                "stunt_type": self.stunt_type,
                "defense_type": self.stunt_type,
                "dt": 3,
            },
        )
        combathandler.msg("$You() prepare a stunt!", self.caller)


class CmdUseItem(_BaseTwitchCombatCommand):
    """
    Use an item in combat. The item must be in your inventory to use.

    Usage:
        use <item>
        use <item> [on] <target>

    Examples:
        use potion
        use throwing knife on goblin
        use bomb goblin

    """

    key = "use"
    help_category = "combat"

    def parse(self):
        super().parse()

        if not self.args:
            self.msg("What do you want to use?")
            raise InterruptCommand()

        self.item = self.lhs
        self.target = self.rhs or "me"

    def func(self):
        item = self.caller.search(
            self.item, candidates=self.caller.equipment.get_usable_objects_from_backpack()
        )
        if not item:
            self.msg("(You must carry the item to use it.)")
            return
        if self.target:
            target = self.caller.search(self.target)
            if not target:
                return

        combathandler = self.get_or_create_combathandler(target)
        combathandler.queue_action({"key": "use", "item": item, "target": target, "dt": 3})
        combathandler.msg(
            f"$You() prepare to use {item.get_display_name(self.caller)}!", self.caller
        )


class CmdWield(_BaseTwitchCombatCommand):
    """
    Wield a weapon or spell-rune. You will the wield the item, swapping with any other item(s) you
    were wielded before.

    Usage:
      wield <weapon or spell>

    Examples:
      wield sword
      wield shield
      wield fireball

    Note that wielding a shield will not replace the sword in your hand, while wielding a two-handed
    weapon (or a spell-rune) will take two hands and swap out what you were carrying.

    """

    key = "wield"
    help_category = "combat"

    def parse(self):
        if not self.args:
            self.msg("What do you want to wield?")
            raise InterruptCommand()
        super().parse()

    def func(self):
        item = self.caller.search(
            self.args, candidates=self.caller.equipment.get_wieldable_objects_from_backpack()
        )
        if not item:
            self.msg("(You must carry the item to wield it.)")
            return
        combathandler = self.get_or_create_combathandler()
        combathandler.queue_action({"key": "wield", "item": item, "dt": 3})
        combathandler.msg(f"$You() reach for {item.get_display_name(self.caller)}!", self.caller)


class TwitchCombatCmdSet(CmdSet):
    """
    Add to character, to be able to attack others in a twitch-style way.
    """

    def at_cmdset_creation(self):
        self.add(CmdAttack())
        self.add(CmdHold())
        self.add(CmdStunt())
        self.add(CmdUseItem())
        self.add(CmdWield())


class TwitchLookCmdSet(CmdSet):
    """
    This will be added/removed dynamically when in combat.
    """

    def at_cmdset_creation(self):
        self.add(CmdLook())



"""
This reroutes from an URL to a python view-function/class.

The main web/urls.py includes these routes for all urls starting with `admin/`
(the `admin/` part should not be included again here).

"""


from django.urls import path

from evennia.web.admin.urls import urlpatterns as evennia_admin_urlpatterns

# add patterns here
urlpatterns = [
    # path("url-pattern", imported_python_view),
    # path("url-pattern", imported_python_view),
]

# read by Django
urlpatterns = urlpatterns + evennia_admin_urlpatterns



"""
SimpleDoor

Contribution - Griatch 2016

A simple two-way exit that represents a door that can be opened and
closed. Can easily be expanded from to make it lockable, destroyable
etc.  Note that the simpledoor is based on Evennia locks, so it will
not work for a superuser (which bypasses all locks) - the superuser
will always appear to be able to close/open the door over and over
without the locks stopping you. To use the door, use `@quell` or a
non-superuser account.

Installation:


Import this module in mygame/commands/default_cmdsets and add
the SimpleDoorCmdSet to your CharacterCmdSet:

```python
# in mygame/commands/default_cmdsets.py

from evennia.contrib.grid import simpledoor  <---

class CharacterCmdSet(default_cmds.CharacterCmdSet):
    # ...
    def at_cmdset_creation(self):
        # ...
        self.add(simpledoor.SimpleDoorCmdSet)

```

Usage:

To try it out, `dig` a new room and then use the (overloaded) `@open`
commmand to open a new doorway to it like this:

    @open doorway:contrib.grid.simpledoor.SimpleDoor = otherroom

You can then use `open doorway' and `close doorway` to change the open
state. If you are not superuser (`@quell` yourself) you'll find you
cannot pass through either side of the door once it's closed from the
other side.

"""

from evennia import DefaultExit, default_cmds
from evennia.commands.cmdset import CmdSet
from evennia.utils.utils import inherits_from


class SimpleDoor(DefaultExit):
    """
    A two-way exit "door" with some methods for affecting both "sides"
    of the door at the same time. For example, set a lock on either of the two
    sides using `exitname.setlock("traverse:false())`

    """

    def at_object_creation(self):
        """
        Called the very first time the door is created.

        """
        self.db.return_exit = None

    def setlock(self, lockstring):
        """
        Sets identical locks on both sides of the door.

        Args:
            lockstring (str): A lockstring, like `"traverse:true()"`.

        """
        self.locks.add(lockstring)
        self.db.return_exit.locks.add(lockstring)

    def setdesc(self, description):
        """
        Sets identical descs on both sides of the door.

        Args:
            setdesc (str): A description.

        """
        self.db.desc = description
        self.db.return_exit.db.desc = description

    def delete(self):
        """
        Deletes both sides of the door.

        """
        # we have to be careful to avoid a delete-loop.
        if self.db.return_exit:
            super().delete()
        super().delete()
        return True

    def at_failed_traverse(self, traversing_object):
        """
        Called when door traverse: lock fails.

        Args:
            traversing_object (Typeclassed entity): The object
                attempting the traversal.

        """
        traversing_object.msg("%s is closed." % self.key)


class CmdOpen(default_cmds.CmdOpen):
    __doc__ = default_cmds.CmdOpen.__doc__
    # overloading parts of the default CmdOpen command to support doors.

    def create_exit(self, exit_name, location, destination, exit_aliases=None, typeclass=None):
        """
        Simple wrapper for the default CmdOpen.create_exit
        """
        # create a new exit as normal
        new_exit = super().create_exit(
            exit_name, location, destination, exit_aliases=exit_aliases, typeclass=typeclass
        )
        if hasattr(self, "return_exit_already_created"):
            # we don't create a return exit if it was already created (because
            # we created a door)
            del self.return_exit_already_created
            return new_exit
        if inherits_from(new_exit, SimpleDoor):
            # a door - create its counterpart and make sure to turn off the default
            # return-exit creation of CmdOpen
            self.caller.msg(
                "Note: A door-type exit was created - ignored eventual custom return-exit type."
            )
            self.return_exit_already_created = True
            back_exit = self.create_exit(
                exit_name, destination, location, exit_aliases=exit_aliases, typeclass=typeclass
            )
            new_exit.db.return_exit = back_exit
            back_exit.db.return_exit = new_exit
        return new_exit


# A simple example of a command making use of the door exit class'
# functionality. One could easily expand it with functionality to
# operate on other types of open-able objects as needed.


class CmdOpenCloseDoor(default_cmds.MuxCommand):
    """
    Open and close a door

    Usage:
        open <door>
        close <door>

    """

    key = "open"
    aliases = ["close"]
    locks = "cmd:all()"
    help_category = "General"

    def func(self):
        "implement the door functionality"
        if not self.args:
            self.caller.msg("Usage: open||close <door>")
            return

        door = self.caller.search(self.args)
        if not door:
            return
        if not inherits_from(door, SimpleDoor):
            self.caller.msg("This is not a door.")
            return

        if self.cmdstring == "open":
            if door.locks.check(self.caller, "traverse"):
                self.caller.msg("%s is already open." % door.key)
            else:
                door.setlock("traverse:true()")
                self.caller.msg("You open %s." % door.key)
        else:  # close
            if not door.locks.check(self.caller, "traverse"):
                self.caller.msg("%s is already closed." % door.key)
            else:
                door.setlock("traverse:false()")
                self.caller.msg("You close %s." % door.key)


class SimpleDoorCmdSet(CmdSet):
    def at_cmdset_creation(self):
        self.add(CmdOpen())
        self.add(CmdOpenCloseDoor())



"""
Helpers for testing evadventure modules.

"""

from evennia.utils import create

from .. import enums
from ..characters import EvAdventureCharacter
from ..objects import (
    EvAdventureArmor,
    EvAdventureHelmet,
    EvAdventureObject,
    EvAdventureShield,
    EvAdventureWeapon,
)
from ..rooms import EvAdventureRoom


class EvAdventureMixin:
    """
    Provides a set of pre-made characters.

    """

    def setUp(self):
        super().setUp()
        self.location = create.create_object(EvAdventureRoom, key="testroom")
        self.character = create.create_object(
            EvAdventureCharacter, key="testchar", location=self.location
        )
        self.helmet = create.create_object(
            EvAdventureHelmet,
            key="helmet",
        )
        self.shield = create.create_object(
            EvAdventureShield,
            key="shield",
        )
        self.armor = create.create_object(
            EvAdventureArmor,
            key="armor",
        )
        self.weapon = create.create_object(
            EvAdventureWeapon,
            key="weapon",
        )
        self.big_weapon = create.create_object(
            EvAdventureWeapon,
            key="big_weapon",
            attributes=[("inventory_use_slot", enums.WieldLocation.TWO_HANDS)],
        )
        self.item = create.create_object(EvAdventureObject, key="backpack item")



"""
This module handles serialization of arbitrary python structural data,
intended primarily to be stored in the database. It also supports
storing Django model instances (which plain pickle cannot do).

This serialization is used internally by the server, notably for
storing data in Attributes and for piping data to process pools.

The purpose of dbserialize is to handle all forms of data. For
well-structured non-arbitrary exchange, such as communicating with a
rich web client, a simpler JSON serialization makes more sense.

This module also implements the `SaverList`, `SaverDict` and `SaverSet`
classes. These are iterables that track their position in a nested
structure and makes sure to send updates up to their root. This is
used by Attributes - without it, one would not be able to update mutables
in-situ, e.g `obj.db.mynestedlist[3][5] = 3` would never be saved and
be out of sync with the database.

"""
from collections import OrderedDict, defaultdict, deque
from collections.abc import MutableMapping, MutableSequence, MutableSet
from functools import update_wrapper

try:
    from pickle import UnpicklingError, dumps, loads
except ImportError:
    from pickle import dumps, loads

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.utils.safestring import SafeString

import evennia
from evennia.utils import logger
from evennia.utils.utils import is_iter, to_bytes, uses_database

__all__ = ("to_pickle", "from_pickle", "do_pickle", "do_unpickle", "dbserialize", "dbunserialize")

PICKLE_PROTOCOL = 2


# message to send if editing an already deleted Attribute in a savermutable
_ERROR_DELETED_ATTR = (
    "{cls_name} {obj} has had its root Attribute deleted. "
    "It must be cast to a {non_saver_name} before it can be modified further."
)


def _get_mysql_db_version():
    """
    This is a helper method for specifically getting the version
    string of a MySQL database.

    Returns:
        mysql_version (str): The currently used mysql database
            version.

    """
    from django.db import connection

    conn = connection.cursor()
    conn.execute("SELECT VERSION()")
    version = conn.fetchone()
    return version and str(version[0]) or ""


# initialization and helpers


_GA = object.__getattribute__
_SA = object.__setattr__
_FROM_MODEL_MAP = None
_TO_MODEL_MAP = None
_IGNORE_DATETIME_MODELS = None


def _IS_PACKED_DBOBJ(o):
    return isinstance(o, tuple) and len(o) == 4 and o[0] == "__packed_dbobj__"


def _IS_PACKED_SESSION(o):
    return isinstance(o, tuple) and len(o) == 3 and o[0] == "__packed_session__"


if uses_database("mysql") and _get_mysql_db_version() < "5.6.4":
    # mysql <5.6.4 don't support millisecond precision
    _DATESTRING = "%Y:%m:%d-%H:%M:%S:000000"
else:
    _DATESTRING = "%Y:%m:%d-%H:%M:%S:%f"


def _TO_DATESTRING(obj):
    """
    Creates datestring hash.

    Args:
        obj (Object): Database object.

    Returns:
        datestring (str): A datestring hash.

    """
    try:
        return _GA(obj, "db_date_created").strftime(_DATESTRING)
    except AttributeError:
        # this can happen if object is not yet saved - no datestring is then set
        try:
            obj.save()
        except AttributeError:
            # we have received a None object, for example due to an erroneous save.
            return None
        return _GA(obj, "db_date_created").strftime(_DATESTRING)


def _init_globals():
    """Lazy importing to avoid circular import issues"""
    global _FROM_MODEL_MAP, _TO_MODEL_MAP, _IGNORE_DATETIME_MODELS
    if not _FROM_MODEL_MAP:
        _FROM_MODEL_MAP = defaultdict(str)
        _FROM_MODEL_MAP.update(dict((c.model, c.natural_key()) for c in ContentType.objects.all()))
    if not _TO_MODEL_MAP:
        from django.conf import settings

        _TO_MODEL_MAP = defaultdict(str)
        _TO_MODEL_MAP.update(
            dict((c.natural_key(), c.model_class()) for c in ContentType.objects.all())
        )
        _IGNORE_DATETIME_MODELS = []
        for src_key, dst_key in settings.ATTRIBUTE_STORED_MODEL_RENAME:
            _TO_MODEL_MAP[src_key] = _TO_MODEL_MAP.get(dst_key, None)
            _IGNORE_DATETIME_MODELS.append(src_key)


#
# SaverList, SaverDict, SaverSet - Attribute-specific helper classes and functions
#


def _save(method):
    """method decorator that saves data to Attribute"""

    def save_wrapper(self, *args, **kwargs):
        self.__doc__ = method.__doc__
        ret = method(self, *args, **kwargs)
        self._save_tree()
        return ret

    return update_wrapper(save_wrapper, method)


class _SaverMutable:
    """
    Parent class for properly handling  of nested mutables in
    an Attribute. If not used something like
     obj.db.mylist[1][2] = "test" (allocation to a nested list)
    will not save the updated value to the database.
    """

    def __init__(self, *args, **kwargs):
        """store all properties for tracking the tree"""
        self._parent = kwargs.pop("_parent", None)
        self._db_obj = kwargs.pop("_db_obj", None)
        self._data = None

    def __bool__(self):
        """Make sure to evaluate as False if empty"""
        return bool(self._data)

    def _save_tree(self):
        """recursively traverse back up the tree, save when we reach the root"""
        if self._parent:
            self._parent._save_tree()
        elif self._db_obj:
            if not self._db_obj.pk:
                cls_name = self.__class__.__name__
                try:
                    non_saver_name = cls_name.split("_Saver", 1)[1].lower()
                except IndexError:
                    non_saver_name = cls_name
                raise ValueError(
                    _ERROR_DELETED_ATTR.format(
                        cls_name=cls_name, obj=self, non_saver_name=non_saver_name
                    )
                )
            self._db_obj.value = self
        else:
            logger.log_err("_SaverMutable %s has no root Attribute to save to." % self)

    def _convert_mutables(self, data):
        """converts mutables to Saver* variants and assigns ._parent property"""

        def process_tree(item, parent):
            """recursively populate the tree, storing parents"""
            dtype = type(item)
            if dtype in (str, int, float, bool, tuple):
                return item
            elif dtype == list:
                dat = _SaverList(_parent=parent)
                dat._data.extend(process_tree(val, dat) for val in item)
                return dat
            elif dtype == dict:
                dat = _SaverDict(_parent=parent)
                dat._data.update((key, process_tree(val, dat)) for key, val in item.items())
                return dat
            elif dtype == defaultdict:
                dat = _SaverDefaultDict(item.default_factory, _parent=parent)
                dat._data.update((key, process_tree(val, dat)) for key, val in item.items())
                return dat
            elif dtype == deque:
                dat = _SaverDeque(_parent=parent, maxlen=item.maxlen)
            elif dtype == set:
                dat = _SaverSet(_parent=parent)
                dat._data.update(process_tree(val, dat) for val in item)
                return dat
            return item

        return process_tree(data, self)

    def __repr__(self):
        return self._data.__repr__()

    def __len__(self):
        return self._data.__len__()

    def __iter__(self):
        return self._data.__iter__()

    def __getitem__(self, key):
        return self._data.__getitem__(key)

    def __eq__(self, other):
        return self._data == other

    def __ne__(self, other):
        return self._data != other

    def __lt__(self, other):
        return self._data < other

    def __gt__(self, other):
        return self._data > other

    def __or__(self, other):
        return self._data | other

    def __ror__(self, other):
        return self._data | other

    @_save
    def __setitem__(self, key, value):
        self._data.__setitem__(key, self._convert_mutables(value))

    @_save
    def __delitem__(self, key):
        self._data.__delitem__(key)

    def deserialize(self):
        """Deserializes this mutable into its corresponding non-Saver type."""
        return deserialize(self)


class _SaverList(_SaverMutable, MutableSequence):
    """
    A list that saves itself to an Attribute when updated.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._data = kwargs.pop("_class", list)()

    @_save
    def __iadd__(self, otherlist):
        self._data = self._data.__add__(otherlist)
        return self._data

    def __add__(self, otherlist):
        return list(self._data) + otherlist

    @_save
    def insert(self, index, value):
        self._data.insert(index, self._convert_mutables(value))

    def __eq__(self, other):
        try:
            return list(self._data) == list(other)
        except TypeError:
            return False

    def __ne__(self, other):
        try:
            return list(self._data) != list(other)
        except TypeError:
            return True

    def index(self, value, *args):
        return self._data.index(value, *args)

    @_save
    def sort(self, *, key=None, reverse=False):
        self._data.sort(key=key, reverse=reverse)

    def copy(self):
        return self._data.copy()


class _SaverDict(_SaverMutable, MutableMapping):
    """
    A dict that stores changes to an Attribute when updated
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._data = kwargs.pop("_class", dict)()

    def has_key(self, key):
        return key in self._data

    @_save
    def update(self, *args, **kwargs):
        self._data.update(*args, **kwargs)


class _SaverDefaultDict(_SaverDict):
    """
    A defaultdict that stores changes to an attribute when updated
    """

    def __init__(self, factory, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._data = defaultdict(factory)
        self.default_factory = factory

    def __getitem__(self, key):
        if key not in self._data.keys():
            # detect the case of db.foo['a'] with no immediate assignment
            # (important: using `key in self._data` would be always True!)
            default_value = self._data[key]
            self.__setitem__(key, default_value)
        return self._data[key]


class _SaverSet(_SaverMutable, MutableSet):
    """
    A set that saves to an Attribute when updated
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._data = set()

    def __contains__(self, value):
        return self._data.__contains__(value)

    @_save
    def add(self, value):
        self._data.add(self._convert_mutables(value))

    @_save
    def discard(self, value):
        self._data.discard(value)

    @_save
    def update(self, *others):
        self._data.update(*others)

    @_save
    def remove(self, value):
        self._data.remove(value)

    @_save
    def pop(self):
        self._data.pop()

    @_save
    def clear(self):
        self._data.clear()

    @_save
    def intersection_update(self, *others):
        self._data.intersection_update(*others)

    @_save
    def difference_update(self, *others):
        self._data.difference_update(*others)

    @_save
    def symmetric_difference_update(self, *others):
        self._data.symmetric_difference(*others)

    def isdisjoint(self, other):
        return self._data.isdisjoint(other)

    def issubset(self, other):
        return self._data.issubset(other)

    def issuperset(self, other):
        return self._data.issuperset(other)

    def union(self, *others):
        return self._data.union(*others)

    def intersection(self, *others):
        return self._data.intersection(*others)

    def difference(self, *others):
        return self._data.difference(*others)

    def symmetric_difference(self, other):
        return self._data.symmetric_difference(other)

    def copy(self):
        return self._data.copy()


class _SaverOrderedDict(_SaverMutable, MutableMapping):
    """
    An ordereddict that can be saved and operated on.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._data = OrderedDict()

    def has_key(self, key):
        return key in self._data


class _SaverDeque(_SaverMutable):
    """
    A deque that can be saved and operated on.
    """

    def __init__(self, *args, maxlen=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._data = deque((), maxlen=maxlen)

    @_save
    def append(self, *args, **kwargs):
        self._data.append(*args, **kwargs)

    @_save
    def appendleft(self, *args, **kwargs):
        self._data.appendleft(*args, **kwargs)

    @_save
    def clear(self):
        self._data.clear()

    @_save
    def extendleft(self, *args, **kwargs):
        self._data.extendleft(*args, **kwargs)

    # maxlen property
    def _getmaxlen(self):
        return self._data.maxlen

    def _setmaxlen(self, value):
        self._data.maxlen = value

    def _delmaxlen(self):
        del self._data.maxlen

    maxlen = property(_getmaxlen, _setmaxlen, _delmaxlen)

    @_save
    def pop(self, *args, **kwargs):
        return self._data.pop(*args, **kwargs)

    @_save
    def popleft(self, *args, **kwargs):
        return self._data.popleft(*args, **kwargs)

    @_save
    def reverse(self):
        self._data.reverse()

    @_save
    def rotate(self, *args):
        self._data.rotate(*args)

    @_save
    def remove(self, *args):
        self._data.remove(*args)


_DESERIALIZE_MAPPING = {
    _SaverList.__name__: list,
    _SaverDict.__name__: dict,
    _SaverSet.__name__: set,
    _SaverOrderedDict.__name__: OrderedDict,
    _SaverDeque.__name__: deque,
    _SaverDefaultDict.__name__: defaultdict,
}


def deserialize(obj):
    """
    Make sure to *fully* decouple a structure from the database, by turning all _Saver*-mutables
    inside it back into their normal Python forms.

    """

    def _iter(obj):
        # breakpoint()
        typ = type(obj)
        tname = typ.__name__
        if tname in ("_SaverDict", "dict"):
            return {_iter(key): _iter(val) for key, val in obj.items()}
        elif tname in ("_SaverOrderedDict", "OrderedDict"):
            return OrderedDict([(_iter(key), _iter(val)) for key, val in obj.items()])
        elif tname in ("_SaverDefaultDict", "defaultdict"):
            return defaultdict(
                obj.default_factory, {_iter(key): _iter(val) for key, val in obj.items()}
            )
        elif tname in ("_SaverDeque", deque):
            return deque((_iter(val) for val in obj), maxlen=obj.maxlen)
        elif tname in _DESERIALIZE_MAPPING:
            return _DESERIALIZE_MAPPING[tname](_iter(val) for val in obj)
        elif is_iter(obj):
            return typ(_iter(val) for val in obj)
        return obj

    return _iter(obj)


#
# serialization helpers


def pack_dbobj(item):
    """
    Check and convert django database objects to an internal representation.

    Args:
        item (any): A database entity to pack

    Returns:
        packed (any or tuple): Either returns the original input item
            or the packing tuple `("__packed_dbobj__", key, creation_time, id)`.

    """
    _init_globals()
    obj = item
    natural_key = _FROM_MODEL_MAP[
        hasattr(obj, "id")
        and hasattr(obj, "db_date_created")
        and hasattr(obj, "__dbclass__")
        and obj.__dbclass__.__name__.lower()
    ]
    # build the internal representation as a tuple
    #  ("__packed_dbobj__", key, creation_time, id)
    return (
        natural_key
        and ("__packed_dbobj__", natural_key, _TO_DATESTRING(obj), _GA(obj, "id"))
        or item
    )


def unpack_dbobj(item):
    """
    Check and convert internal representations back to Django database
    models.

    Args:
        item (packed_dbobj): The fact that item is a packed dbobj
            should be checked before this call.

    Returns:
        unpacked (any): Either the original input or converts the
            internal store back to a database representation (its
            typeclass is returned if applicable).

    """
    _init_globals()
    try:
        obj = item[3] and _TO_MODEL_MAP[item[1]].objects.get(id=item[3])
    except ObjectDoesNotExist:
        return None
    except TypeError:
        if hasattr(item, "pk"):
            # this happens if item is already an obj
            return item
        return None
    if item[1] in _IGNORE_DATETIME_MODELS:
        # if we are replacing models we ignore the datatime
        return obj
    else:
        # even if we got back a match, check the sanity of the date (some
        # databases may 're-use' the id)
        return _TO_DATESTRING(obj) == item[2] and obj or None


def pack_session(item):
    """
    Handle the safe serializion of Sessions objects (these contain
    hidden references to database objects (accounts, puppets) so they
    can't be safely serialized).

    Args:
        item (Session)): This item must have all properties of a session
            before entering this call.

    Returns:
        packed (tuple or None): A session-packed tuple on the form
            `(__packed_session__, sessid, conn_time)`. If this sessid
            does not match a session in the Session handler, None is returned.

    """
    _init_globals()
    session = evennia.SESSION_HANDLER.get(item.sessid)
    if session and session.conn_time == item.conn_time:
        # we require connection times to be identical for the Session
        # to be accepted as actually being a session (sessids gets
        # reused all the time).
        return (
            item.conn_time
            and item.sessid
            and ("__packed_session__", _GA(item, "sessid"), _GA(item, "conn_time"))
        )
    return None


def unpack_session(item):
    """
    Check and convert internal representations back to Sessions.

    Args:
        item (packed_session): The fact that item is a packed session
            should be checked before this call.

    Returns:
        unpacked (any): Either the original input or converts the
            internal store back to a Session. If Session no longer
            exists, None will be returned.
    """
    _init_globals()
    session = evennia.SESSION_HANDLER.get(item[1])
    if session and session.conn_time == item[2]:
        # we require connection times to be identical for the Session
        # to be accepted as the same as the one stored (sessids gets
        # reused all the time).
        return session
    return None


#
# Access methods


def to_pickle(data):
    """
    This prepares data on arbitrary form to be pickled. It handles any
    nested structure and returns data on a form that is safe to pickle
    (including having converted any database models to their internal
    representation).  We also convert any Saver*-type objects back to
    their normal representations, they are not pickle-safe.

    Args:
        data (any): Data to pickle.

    Returns:
        data (any): Pickled data.

    """

    def process_item(item):
        """Recursive processor and identification of data"""

        dtype = type(item)

        if dtype in (str, int, float, bool, bytes, SafeString):
            return item
        elif dtype == tuple:
            return tuple(process_item(val) for val in item)
        elif dtype in (list, _SaverList):
            return [process_item(val) for val in item]
        elif dtype in (dict, _SaverDict):
            return dict((process_item(key), process_item(val)) for key, val in item.items())
        elif dtype in (defaultdict, _SaverDefaultDict):
            return defaultdict(
                item.default_factory,
                ((process_item(key), process_item(val)) for key, val in item.items()),
            )
        elif dtype in (deque, _SaverDeque):
            return deque((process_item(val) for val in item), maxlen=item.maxlen)
        elif dtype in (set, _SaverSet):
            return set(process_item(val) for val in item)
        elif dtype in (OrderedDict, _SaverOrderedDict):
            return OrderedDict((process_item(key), process_item(val)) for key, val in item.items())
        elif dtype in (deque, _SaverDeque):
            return deque(process_item(val) for val in item)

        # not one of the base types
        if hasattr(item, "__serialize_dbobjs__"):
            # Allows custom serialization of any dbobjects embedded in
            # the item that Evennia will otherwise not find (these would
            # otherwise lead to an error). Use the dbserialize helper from
            # this method.
            try:
                item.__serialize_dbobjs__()
            except TypeError as err:
                # we catch typerrors so we can handle both classes (requiring
                # classmethods) and instances
                pass

        if hasattr(item, "__iter__"):
            try:
                # we try to conserve the iterable class, if not convert to dict
                try:
                    return item.__class__(
                        (process_item(key), process_item(val)) for key, val in item.items()
                    )
                except (AttributeError, TypeError):
                    return {process_item(key): process_item(val) for key, val in item.items()}
            except Exception:
                # we try to conserve the iterable class, if not convert to list
                try:
                    return item.__class__([process_item(val) for val in item])
                except (AttributeError, TypeError):
                    return [process_item(val) for val in item]
        elif hasattr(item, "sessid") and hasattr(item, "conn_time"):
            return pack_session(item)
        try:
            return pack_dbobj(item)
        except TypeError:
            return item
        except Exception:
            logger.log_err(f"The object {item} of type {type(item)} could not be stored.")
            raise

    return process_item(data)


# @transaction.autocommit
def from_pickle(data, db_obj=None):
    """
    This should be fed a just de-pickled data object. It will be converted back
    to a form that may contain database objects again. Note that if a database
    object was removed (or changed in-place) in the database, None will be
    returned.

    Args:
        data (any): Pickled data to unpickle.
        db_obj (Atribute, any): This is the model instance (normally
            an Attribute) that _Saver*-type iterables (_SaverList etc)
            will save to when they update. It must have a 'value' property
            that saves assigned data to the database. Skip if not
            serializing onto a given object.  If db_obj is given, this
            function will convert lists, dicts and sets to their
            _SaverList, _SaverDict and _SaverSet counterparts.

    Returns:
        data (any): Unpickled data.

    """

    def process_item(item):
        """Recursive processor and identification of data"""
        # breakpoint()
        dtype = type(item)
        if dtype in (str, int, float, bool, bytes, SafeString):
            return item
        elif _IS_PACKED_DBOBJ(item):
            # this must be checked before tuple
            return unpack_dbobj(item)
        elif _IS_PACKED_SESSION(item):
            return unpack_session(item)
        elif dtype == tuple:
            return tuple(process_item(val) for val in item)
        elif dtype == dict:
            return dict((process_item(key), process_item(val)) for key, val in item.items())
        elif dtype == defaultdict:
            return defaultdict(
                item.default_factory,
                ((process_item(key), process_item(val)) for key, val in item.items()),
            )
        elif dtype == set:
            return set(process_item(val) for val in item)
        elif dtype == OrderedDict:
            return OrderedDict((process_item(key), process_item(val)) for key, val in item.items())
        elif dtype == deque:
            return deque((process_item(val) for val in item), maxlen=item.maxlen)
        elif hasattr(item, "__iter__"):
            try:
                # we try to conserve the iterable class, if not convert to dict
                try:
                    return item.__class__(
                        (process_item(key), process_item(val)) for key, val in item.items()
                    )
                except (AttributeError, TypeError):
                    return {process_item(key): process_item(val) for key, val in item.items()}
            except Exception:
                try:
                    # we try to conserve the iterable class if
                    # it accepts an iterator
                    return item.__class__(process_item(val) for val in item)
                except (AttributeError, TypeError):
                    return [process_item(val) for val in item]

        if hasattr(item, "__deserialize_dbobjs__"):
            # this allows the object to custom-deserialize any embedded dbobjs
            # that we previously serialized with __serialize_dbobjs__.
            # use the dbunserialize helper in this module.
            try:
                item.__deserialize_dbobjs__()
            except (TypeError, UnpicklingError):
                # handle recoveries both of classes (requiring classmethods
                # or instances. Unpickling errors can happen when re-loading the
                # data from cache (because the hidden entity was already
                # deserialized and stored back on the object, unpickling it
                # again fails). TODO: Maybe one could avoid this retry in a
                # more graceful way?
                pass

        return item

    def process_tree(item, parent):
        """Recursive processor, building a parent-tree from iterable data"""
        # breakpoint()
        dtype = type(item)
        if dtype in (str, int, float, bool, bytes, SafeString):
            return item
        elif _IS_PACKED_DBOBJ(item):
            # this must be checked before tuple
            return unpack_dbobj(item)
        elif dtype == tuple:
            return tuple(process_tree(val, item) for val in item)
        elif dtype == list:
            dat = _SaverList(_parent=parent)
            dat._data.extend(process_tree(val, dat) for val in item)
            return dat
        elif dtype == dict:
            dat = _SaverDict(_parent=parent)
            dat._data.update(
                (process_item(key), process_tree(val, dat)) for key, val in item.items()
            )
            return dat
        elif dtype == defaultdict:
            dat = _SaverDefaultDict(item.default_factory, _parent=parent)
            dat._data.update(
                (process_item(key), process_tree(val, dat)) for key, val in item.items()
            )
            return dat
        elif dtype == set:
            dat = _SaverSet(_parent=parent)
            dat._data.update(set(process_tree(val, dat) for val in item))
            return dat
        elif dtype == OrderedDict:
            dat = _SaverOrderedDict(_parent=parent)
            dat._data.update(
                (process_item(key), process_tree(val, dat)) for key, val in item.items()
            )
            return dat
        elif dtype == deque:
            dat = _SaverDeque(_parent=parent, maxlen=item.maxlen)
            dat._data.extend(process_item(val) for val in item)
            return dat
        elif hasattr(item, "__iter__"):
            try:
                # we try to conserve the iterable class, if not convert to dict
                try:
                    dat = _SaverDict(_parent=parent, _class=item.__class__)
                    dat._data.update(
                        (process_item(key), process_tree(val, dat)) for key, val in item.items()
                    )
                    return dat
                except (AttributeError, TypeError):
                    dat = _SaverDict(_parent=parent)
                    dat._data.update(
                        (process_item(key), process_tree(val, dat)) for key, val in item.items()
                    )
                    return dat
            except Exception:
                try:
                    # we try to conserve the iterable class if it
                    # accepts an iterator
                    dat = _SaverList(_parent=parent, _class=item.__class__)
                    dat._data.extend(process_tree(val, dat) for val in item)
                    return dat
                except (AttributeError, TypeError):
                    dat = _SaverList(_parent=parent)
                    dat._data.extend(process_tree(val, dat) for val in item)
                    return dat

        if hasattr(item, "__deserialize_dbobjs__"):
            try:
                item.__deserialize_dbobjs__()
            except (TypeError, UnpicklingError):
                pass

        return item

    if db_obj:
        # convert lists, dicts and sets to their Saved* counterparts. It
        # is only relevant if the "root" is an iterable of the right type.
        dtype = type(data)
        if dtype in (str, int, float, bool, bytes, SafeString, tuple):
            return process_item(data)
        elif dtype == list:
            dat = _SaverList(_db_obj=db_obj)
            dat._data.extend(process_tree(val, dat) for val in data)
            return dat
        elif dtype == dict:
            dat = _SaverDict(_db_obj=db_obj)
            dat._data.update(
                (process_item(key), process_tree(val, dat)) for key, val in data.items()
            )
            return dat
        elif dtype == defaultdict:
            dat = _SaverDefaultDict(data.default_factory, _db_obj=db_obj)
            dat._data.update(
                (process_item(key), process_tree(val, dat)) for key, val in data.items()
            )
            return dat
        elif dtype == set:
            dat = _SaverSet(_db_obj=db_obj)
            dat._data.update(process_tree(val, dat) for val in data)
            return dat
        elif dtype == OrderedDict:
            dat = _SaverOrderedDict(_db_obj=db_obj)
            dat._data.update(
                (process_item(key), process_tree(val, dat)) for key, val in data.items()
            )
            return dat
        elif dtype == deque:
            dat = _SaverDeque(_db_obj=db_obj, maxlen=data.maxlen)
            dat._data.extend(process_item(val) for val in data)
            return dat
        elif hasattr(data, "__iter__"):
            try:
                # we try to conserve the iterable class, if not convert to dict
                try:
                    dat = _SaverDict(_db_obj=db_obj, _class=data.__class__)
                    dat._data.update(
                        (process_item(key), process_tree(val, dat)) for key, val in data.items()
                    )
                    return dat
                except (AttributeError, TypeError):
                    dat = _SaverDict(_db_obj=db_obj)
                    dat._data.update(
                        (process_item(key), process_tree(val, dat)) for key, val in data.items()
                    )
                    return dat
            except Exception:
                try:
                    # we try to conserve the iterable class if it
                    # accepts an iterator
                    dat = _SaverList(_db_obj=db_obj, _class=data.__class__)
                    dat._data.extend(process_tree(val, dat) for val in data)
                    return dat

                except (AttributeError, TypeError):
                    dat = _SaverList(_db_obj=db_obj)
                    dat._data.extend(process_tree(val, dat) for val in data)
                    return dat

    return process_item(data)


def do_pickle(data):
    """Perform pickle to string"""
    try:
        return dumps(data, protocol=PICKLE_PROTOCOL)
    except Exception:
        logger.log_err(f"Could not pickle data for storage: {data}")
        raise


def do_unpickle(data):
    """Retrieve pickle from pickled string"""
    try:
        return loads(to_bytes(data))
    except Exception:
        logger.log_err(f"Could not unpickle data from storage: {data}")
        raise


def dbserialize(data):
    """Serialize to pickled form in one step"""
    return do_pickle(to_pickle(data))


def dbunserialize(data, db_obj=None):
    """Un-serialize in one step. See from_pickle for help db_obj."""
    return from_pickle(do_unpickle(data), db_obj=db_obj)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-09-25 17:35


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("comms", "0015_auto_20170706_2041")]

    operations = [
        migrations.RemoveField(model_name="channeldb", name="db_subscriptions"),
        migrations.AlterField(
            model_name="msg", name="db_message", field=models.TextField(verbose_name="message")
        ),
    ]



"""

NAWS - Negotiate About Window Size

This implements the NAWS telnet option as per
https://www.ietf.org/rfc/rfc1073.txt

NAWS allows telnet clients to report their current window size to the
client and update it when the size changes

"""
from codecs import encode as codecs_encode

from django.conf import settings

NAWS = bytes([31])  # b"\x1f"
IS = bytes([0])  # b"\x00"

# default taken from telnet specification
DEFAULT_WIDTH = settings.CLIENT_DEFAULT_WIDTH
DEFAULT_HEIGHT = settings.CLIENT_DEFAULT_HEIGHT

# try to get the customized mssp info, if it exists.


class Naws:
    """
    Implements the NAWS protocol. Add this to a variable on the telnet
    protocol to set it up.

    """

    def __init__(self, protocol):
        """
        initialize NAWS by storing protocol on ourselves and calling
        the client to see if it supports NAWS.

        Args:
            protocol (Protocol): The active protocol instance.

        """
        self.naws_step = 0
        self.protocol = protocol
        self.protocol.protocol_flags["SCREENWIDTH"] = {
            0: DEFAULT_WIDTH
        }  # windowID (0 is root):width
        self.protocol.protocol_flags["SCREENHEIGHT"] = {0: DEFAULT_HEIGHT}  # windowID:width
        self.protocol.negotiationMap[NAWS] = self.negotiate_sizes
        self.protocol.do(NAWS).addCallbacks(self.do_naws, self.no_naws)

    def no_naws(self, option):
        """
        Called when client is not reporting NAWS. This is the normal
        operation.

        Args:
            option (Option): Not used.

        """
        self.protocol.handshake_done()

    def do_naws(self, option):
        """
        Client wants to negotiate all the NAWS information.

        Args:
            option (Option): Not used.

        """
        self.protocol.handshake_done()

    def negotiate_sizes(self, options):
        """
        Step through the NAWS handshake.

        Args:
            option (list): The incoming NAWS options.

        """
        if len(options) == 4:
            # NAWS is negotiated with 16bit words
            width = options[0] + options[1]
            self.protocol.protocol_flags["SCREENWIDTH"][0] = int(codecs_encode(width, "hex"), 16)
            height = options[2] + options[3]
            self.protocol.protocol_flags["SCREENHEIGHT"][0] = int(codecs_encode(height, "hex"), 16)



"""
Test menu_login

"""

from evennia.commands.default.tests import BaseEvenniaCommandTest

from . import menu_login


class TestMenuLogin(BaseEvenniaCommandTest):
    def test_cmdunloggedlook(self):
        self.call(menu_login.CmdUnloggedinLook(), "", "======")



"""
An example EvMenu for the character creator contrib.

This menu is not intended to be a full character creator, but to demonstrate
several different useful node types for your own creator. Any of the different
techniques demonstrated here can be combined into a single decision point.

## Informational Pages

A set of nodes that let you page through information on different choices.

The example shows how to have a single informational page for each option, but
you can expand that into sub-categories by setting the values to dictionaries
instead of simple strings and referencing the "Option Categories" nodes.

## Option Categories

A pair of nodes which let you divide your options into separate categories.
The base node has a list of categories as the options, which is passed to the
child node. That child node then presents the options for that category and
allows the player to choose one.

## Multiple Choice

Allows players to select and deselect options from the list in order to choose
more than one. The example has a requirement of choosing exactly 3 options,
but you can change it to a maximum or minimum number of required options -
or remove the requirement check entirely.

## Simple List Options

If you just want a straightforward list of options, without any of the back-and-forth
navigation or modifying of option text, evennia has an  easy to use decorator
available: `@list_node`

For an example of how to use it, check out the documentation for evennia.utils.evmenu
- there's lots of other useful EvMenu tools too!

## Starting Objects

Allows players to choose from a selection of starting objects.

When creating starting objects e.g. gear, it's best to actually create them
at the end, so you don't have to patch checks in for orphaned objects or
infinite-object player exploits.

## Choosing a Name

The contrib character create command assumes the player will choose their name
during character creation. This section validates name choices before confirming
them.

## The End

It might not seem like an important part, since the players don't make a decision
here, but the final node of character creation is where you finalize all of
the decisions players made earlier. Initializing skills, creating starting gear,
and other one-time method calls and set-up should be put here.
"""

import inflect
from typeclasses.characters import Character

from evennia.prototypes.spawner import spawn
from evennia.utils import dedent
from evennia.utils.evtable import EvTable

_INFLECT = inflect.engine()


#########################################################
#                   Welcome Page
#########################################################


def menunode_welcome(caller):
    """Starting page."""
    text = dedent(
        """\
        |wWelcome to Character Creation!|n

        This is the starting node for all brand new characters. It's a good place to
        remind players that they can exit the character creator and resume later,
        especially if you're going to have a really long chargen process.

        A brief overview of the game could be a good idea here, too, or a link to your
        game wiki if you have one.
    """
    )
    help = "You can explain the commands for exiting and resuming more specifically here."
    options = {"desc": "Let's begin!", "goto": "menunode_info_base"}
    return (text, help), options


#########################################################
#                 Informational Pages
#########################################################

# Storing your information in a dictionary like this makes the menu nodes much cleaner,
# as well as making info easier to update. You can even import it from a different module,
# e.g. wherever you have the classes actually defined, so later updates only happen in one place.
_CLASS_INFO_DICT = {
    # The keys here are the different options you can choose, and the values are the info pages
    "warrior": dedent(
        """\
        Most warriors specialize in melee weapons, although ranged combat
        is not unheard of.

        Warriors like to compete by beating each other up for fun.
        """
    ),
    "mage": dedent(
        """\
        Mages prefer less combative lines of work, such as showmanship or
        selling enchanted charms. Those who choose to be a battle mage are
        highly sought after by adventuring parties.

        Mage schools, being led by the most academic-minded of mages, are
        notorious for intellectual snobbery.
        """
    ),
}


def menunode_info_base(caller):
    """Base node for the informational choices."""
    # this is a base node for a decision, so we want to save the character's progress here
    caller.new_char.db.chargen_step = "menunode_info_base"

    text = dedent(
        """\
        |wInformational Pages|n

        Sometimes you'll want to let players read more about options before choosing
        one of them. This is especially useful for big choices like race or class.
    """
    )
    help = "A link to your wiki for more information on classes could be useful here."
    options = []
    # Build your options from your info dict so you don't need to update this to add new options
    for pclass in _CLASS_INFO_DICT.keys():
        options.append(
            {
                "desc": f"Learn about the |c{pclass}|n class",
                "goto": ("menunode_info_class", {"selected_class": pclass}),
            }
        )
    return (text, help), options


# putting your kwarg in the menu declaration helps keep track of what variables the node needs
def menunode_info_class(caller, raw_string, selected_class=None, **kwargs):
    """Informational overview of a particular class"""

    # sometimes weird bugs happen - it's best to check for them rather than let the game break
    if not selected_class:
        # reset back to the previous step
        caller.new_char.db.chargen_step = "menunode_welcome"
        # print error to player and quit the menu
        return "Something went wrong. Please try again."

    # Since you have all the info in a nice dict, you can just grab it to display here
    text = _CLASS_INFO_DICT[selected_class]
    help = "If you want option-specific help, you can define it in your info dict and reference it."
    options = []

    # set an option for players to choose this class
    options.append(
        {
            "desc": f"Become {_INFLECT.an(selected_class)}",
            "goto": (_set_class, {"selected_class": selected_class}),
        }
    )

    # once again build your options from the same info dict
    for pclass in _CLASS_INFO_DICT.keys():
        # make sure you don't print the currently displayed page as an option
        if pclass != selected_class:
            options.append(
                {
                    "desc": f"Learn about the |c{pclass}|n class",
                    "goto": ("menunode_info_class", {"selected_class": pclass}),
                }
            )
    return (text, help), options


def _set_class(caller, raw_string, selected_class=None, **kwargs):
    # a class should always be selected here
    if not selected_class:
        # go back to the base node for this decision
        return "menunode_info_base"

    char = caller.new_char
    # any special code for setting this option would go here!
    # but we'll just set an attribute
    char.db.player_class = selected_class

    # move on to the next step!
    return "menunode_categories"


#########################################################
#                Option Categories
#########################################################

# for these subcategory options, we make a dict of categories and option lists
_APPEARANCE_DICT = {
    # the key is your category; the value is a list of options, in the order you want them to appear
    "body type": [
        "skeletal",
        "skinny",
        "slender",
        "slim",
        "athletic",
        "muscular",
        "broad",
        "round",
        "curvy",
        "stout",
        "chubby",
    ],
    "height": ["diminutive", "short", "average", "tall", "towering"],
}


def menunode_categories(caller, **kwargs):
    """Base node for categorized options."""
    # this is a new decision step, so save your resume point here
    caller.new_char.db.chargen_step = "menunode_categories"

    text = dedent(
        """\
        |wOption Categories|n

        Some character attributes are part of the same mechanic or decision,
        but need to be divided up into sub-categories. Character appearance
        details are an example of where this can be useful.
        """
    )

    help = "Some helpful extra information on what's affected by these choices works well here."
    options = []

    # just like for informational categories, build the options off of a dictionary to make it
    # easier to manage
    for category in _APPEARANCE_DICT.keys():
        options.append(
            {
                "desc": f"Choose your |c{category}|n",
                "goto": ("menunode_category_options", {"category": category}),
            }
        )

    # since this node goes in and out of sub-nodes, you need an option to proceed to the next step
    options.append(
        {
            "key": ("(Next)", "next", "n"),
            "desc": "Continue to the next step.",
            "goto": "menunode_multi_choice",
        }
    )
    # once past the first decision, it's also a good idea to include a "back to previous step"
    # option
    options.append(
        {
            "key": ("(Back)", "back", "b"),
            "desc": "Go back to the previous step",
            "goto": "menunode_info_base",
        }
    )
    return (text, help), options


def menunode_category_options(caller, raw_string, category=None, **kwargs):
    """Choosing an option within the categories."""
    if not category:
        # this shouldn't have happened, so quit and retry
        return "Something went wrong. Please try again."

    # for mechanics-related choices, you can combine this with the
    # informational options approach to give specific info
    text = f"Choose your {category}:"
    help = f"This will define your {category}."

    options = []
    # build the list of options from the right category of your dictionary
    for option in _APPEARANCE_DICT[category]:
        options.append(
            {"desc": option, "goto": (_set_category_opt, {"category": category, "value": option})}
        )
    # always include a "back" option in case they aren't ready to pick yet
    options.append(
        {
            "key": ("(Back)", "back", "b"),
            "desc": f"Don't change {category}",
            "goto": "menunode_categories",
        }
    )
    return (text, help), options


def _set_category_opt(caller, raw_string, category, value, **kwargs):
    """Set the option for a category"""

    # this is where you would put any more complex code involved in setting the option,
    # but we're just doing simple attributes
    caller.new_char.attributes.add(category, value)

    # go back to the base node for the categories choice to pick another
    return "menunode_categories"


#########################################################
#                  Multiple Choice
#########################################################

# it's not as important to make this a separate list, but like all the others,
# it's easier to read and to update if you do!
_SKILL_OPTIONS = [
    "alchemy",
    "archery",
    "blacksmithing",
    "brawling",
    "dancing",
    "fencing",
    "pottery",
    "tailoring",
    "weaving",
]


def menunode_multi_choice(caller, raw_string, **kwargs):
    """A multiple-choice menu node."""
    char = caller.new_char

    # another decision, so save the resume point
    char.db.chargen_step = "menunode_multi_choice"

    # in order to support picking up from where we left off, get the options from the character
    # if they weren't passed in
    # this is again just a simple attribute, but you could retrieve this list however
    selected = kwargs.get("selected") or char.attributes.get("skill_list", [])

    text = dedent(
        """\
        |wMultiple Choice|n

        Sometimes you want players to be able to pick more than one option -
        for example, starting skills.

        You can easily define it as a minimum, maximum, or exact number of
        selected options.
    """
    )

    help = (
        "This is a good place to specify how many choices are allowed or required. This example"
        " requires exactly 3."
    )

    options = []
    for option in _SKILL_OPTIONS:
        # check if the option has been selected
        if option in selected:
            # if it's been selected, we want to highlight that
            opt_desc = f"|y{option} (selected)|n"
        else:
            opt_desc = option
        options.append(
            {"desc": opt_desc, "goto": (_set_multichoice, {"selected": selected, "option": option})}
        )

    # only display the Next option if the requirements are met!
    # for this example, you need exactly 3 choices, but you can use an inequality
    # for "no more than X", or "at least X"
    if len(selected) == 3:
        options.append(
            {
                "key": ("(Next)", "next", "n"),
                "desc": "Continue to the next step",
                "goto": "menunode_choose_objects",
            }
        )
    options.append(
        {
            "key": ("(Back)", "back", "b"),
            "desc": "Go back to the previous step",
            "goto": "menunode_categories",
        }
    )

    return (text, help), options


def _set_multichoice(caller, raw_string, selected=[], **kwargs):
    """saves the current choices to the character"""
    # get the option being chosen
    if option := kwargs.get("option"):
        # if the option is already in the list, then we want to remove it
        if option in selected:
            selected.remove(option)
        # otherwise, we're adding it
        else:
            selected.append(option)

        # now that the options are updated, save it to the character
        # this is just setting an attribute but it could be anything
        caller.new_char.attributes.add("skill_list", selected)

    # pass the list back so we don't need to retrieve it again
    return ("menunode_multi_choice", {"selected": selected})


#########################################################
#                  Starting Objects
#########################################################

# for a real game, you would most likely want to build this list from
# your existing game prototypes module(s), but for a demo this'll do!
_EXAMPLE_PROTOTYPES = [
    # starter sword prototype
    {
        "key": "basic sword",
        "desc": "This sword will do fine for stabbing things.",
        "tags": [("sword", "weapon")],
    },
    # starter staff prototype
    {
        "key": "basic staff",
        "desc": "You could hit things with it, or maybe use it as a spell focus.",
        "tags": [("staff", "weapon"), ("staff", "focus")],
    },
]


# this method will be run to create the starting objects
def create_objects(character):
    """do the actual object spawning"""
    # since our example chargen saves the starting prototype to an attribute, we retrieve that here
    proto = dict(character.db.starter_weapon)
    # set the location to our character, so they actually have it
    proto["location"] = character
    # create the object
    spawn(proto)


def menunode_choose_objects(caller, raw_string, **kwargs):
    """Selecting objects to start with"""
    char = caller.new_char

    # another decision, so save the resume point
    char.db.chargen_step = "menunode_choose_objects"

    text = dedent(
        """\
        |wStarting Objects|n

        Whether it's a cosmetic outfit, a starting weapon, or a professional
        tool kit, you probably want to let your players have a choice in
        what objects they start out with.
        """
    )

    help = (
        "An overview of what the choice affects - whether it's purely aesthetic or mechanical, and"
        " whether you can change it later - are good here."
    )

    options = []

    for proto in _EXAMPLE_PROTOTYPES:
        # use the key as the option description, but pass the whole prototype
        options.append(
            {
                "desc": f"Choose {_INFLECT.an(proto['key'])}",
                "goto": (_set_object_choice, {"proto": proto}),
            }
        )

    options.append(
        {
            "key": ("(Back)", "back", "b"),
            "desc": "Go back to the previous step",
            "goto": "menunode_multi_choice",
        }
    )

    return (text, help), options


def _set_object_choice(caller, raw_string, proto, **kwargs):
    """Save the selected starting object(s)"""

    # we DON'T want to actually create the object, yet! that way players can still go back and
    # change their mind instead, we save what object was chosen - in this case, by saving the
    # prototype dict to the character
    caller.new_char.db.starter_weapon = proto

    # continue to the next step
    return "menunode_choose_name"


#########################################################
#                Choosing a Name
#########################################################


def menunode_choose_name(caller, raw_string, **kwargs):
    """Name selection"""
    char = caller.new_char

    # another decision, so save the resume point
    char.db.chargen_step = "menunode_choose_name"

    # check if an error message was passed to the node. if so, you'll want to include it
    # into your "name prompt" at the end of the node text.
    if error := kwargs.get("error"):
        prompt_text = f"{error}. Enter a different name."
    else:
        # there was no error, so just ask them to enter a name.
        prompt_text = "Enter a name here to check if it's available."

    # this will print every time the player is prompted to choose a name,
    # including the prompt text defined above
    text = dedent(
        f"""\
        |wChoosing a Name|n

        Especially for roleplaying-centric games, being able to choose your
        character's name after deciding everything else, instead of before,
        is really useful.

        {prompt_text}
        """
    )

    help = "You'll have a chance to change your mind before confirming, even if the name is free."
    # since this is a free-text field, we just have the one
    options = {"key": "_default", "goto": _check_charname}
    return (text, help), options


def _check_charname(caller, raw_string, **kwargs):
    """Check and confirm name choice"""

    # strip any extraneous whitespace from the raw text
    # if you want to do any other validation on the name, e.g. no punctuation allowed, this
    # is the place!
    charname = raw_string.strip()

    # aside from validation, the built-in normalization function from the caller's Account does
    # some useful cleanup on the input, just in case they try something sneaky
    charname = caller.account.normalize_username(charname)

    # check to make sure that the name doesn't already exist
    candidates = Character.objects.filter_family(db_key__iexact=charname)
    if len(candidates):
        # the name is already taken - report back with the error
        return (
            "menunode_choose_name",
            {"error": f"|w{charname}|n is unavailable.\n\nEnter a different name."},
        )
    else:
        # it's free! set the character's key to the name to reserve it
        caller.new_char.key = charname
        # continue on to the confirmation node
        return "menunode_confirm_name"


def menunode_confirm_name(caller, raw_string, **kwargs):
    """Confirm the name choice"""
    char = caller.new_char

    # since we reserved the name by assigning it, you can reference the character key
    # if you have any extra validation or normalization that changed the player's input
    # this also serves to show the player exactly what name they'll get
    text = f"|w{char.key}|n is available! Confirm?"
    # let players change their mind and go back to the name choice, if they want
    options = [
        {"key": ("Yes", "y"), "goto": "menunode_end"},
        {"key": ("No", "n"), "goto": "menunode_choose_name"},
    ]
    return text, options


#########################################################
#                     The End
#########################################################


def menunode_end(caller, raw_string):
    """End-of-chargen cleanup."""
    char = caller.new_char
    # since everything is finished and confirmed, we actually create the starting objects now
    create_objects(char)

    # clear in-progress status
    caller.new_char.attributes.remove("chargen_step")
    text = dedent(
        """
        Congratulations!

        You have completed character creation. Enjoy the game!
    """
    )
    return text, None



"""Tests for validatorfuncs """

import datetime

import mock
import pytz
from django.test import TestCase

from evennia.utils import validatorfuncs


class TestValidatorFuncs(TestCase):
    def test_text_ok(self):
        for val in [None, -123, "abc", 1.234, {1: True, 2: False}, ["a", 1]]:
            self.assertEqual(str(val), validatorfuncs.text(val))

    @mock.patch("builtins.str")
    def test_text_raises_ValueError(self, mocked_str):
        mocked_str.side_effect = Exception
        with self.assertRaises(ValueError):
            validatorfuncs.text(None)

    def test_color_ok(self):
        for color in ["r", "g", "b", "H", "R", "M", "^"]:
            self.assertEqual(color, validatorfuncs.color(color))

    def test_color_falsy_raises_ValueError(self):
        for color in [None, (), [], False, True, {}]:
            with self.assertRaises(ValueError):
                validatorfuncs.color(color)

    def test_datetime_ok(self):
        for dt in ["Oct 12 1:00 1492", "Jan 2 12:00 2020", "Dec 31 00:00 2018"]:
            self.assertTrue(
                isinstance(validatorfuncs.datetime(dt, from_tz=pytz.UTC), datetime.datetime)
            )
        account = mock.MagicMock()
        account.options.get = mock.MagicMock(return_value="America/Chicago")
        expected = datetime.datetime(1492, 10, 12, 6, 51, tzinfo=pytz.UTC)
        self.assertEqual(expected, validatorfuncs.datetime("Oct 12 1:00 1492", account=account))
        account.options.get.assert_called_with("timezone", "UTC")

    def test_datetime_raises_ValueError(self):
        for dt in ["", "January 1, 2019", "1/1/2019", "Jan 1 2019"]:
            with self.assertRaises(ValueError):
                validatorfuncs.datetime(dt)

    def test_duration_ok(self):
        for d in ["1d", "2w", "3h", "4s", "5m", "6y"]:
            self.assertTrue(isinstance(validatorfuncs.duration(d), datetime.timedelta))

        self.assertEqual(
            datetime.timedelta(1 + 6 * 365, 2, 0, 0, 3, 4, 5),
            validatorfuncs.duration("1d 2s 3m 4h 5w 6y"),
        )
        # values may be duplicated
        self.assertEqual(
            datetime.timedelta((1 + 7) + (6 + 12) * 365, 2 + 8, 0, 0, 3 + 9, 4 + 10, 5 + 11),
            validatorfuncs.duration("1d 2s 3m 4h 5w 6y 7d 8s 9m 10h 11w 12y"),
        )

    def test_duration_raises_ValueError(self):
        for d in ["", "1", "5days", "1Week"]:
            with self.assertRaises(ValueError):
                validatorfuncs.duration(d)

    def test_future_ok(self):
        year = int(datetime.datetime.utcnow().strftime("%Y"))
        for f in [f"Jan 2 12:00 {year+1}", f"Dec 31 00:00 {year+1}"]:
            self.assertTrue(
                isinstance(validatorfuncs.future(f, from_tz=pytz.UTC), datetime.datetime)
            )

    def test_future_raises_ValueError(self):
        year = int(datetime.datetime.utcnow().strftime("%Y"))
        for f in [f"Jan 2 12:00 {year-1}", f"Dec 31 00:00 {year-1}"]:
            with self.assertRaises(ValueError):
                validatorfuncs.future(f, from_tz=pytz.UTC)

    def test_signed_integer_ok(self):
        for si in ["123", "4567890", "001", "-123", "-45", "0"]:
            self.assertEqual(int(si), validatorfuncs.signed_integer(si))

    @mock.patch("builtins.int")
    def test_signed_integer_raises_ValueError(self, mocked_int):
        for si in ["", "000", "abc"]:
            mocked_int.side_effect = ValueError
            with self.assertRaises(ValueError):
                validatorfuncs.signed_integer(si)

    def test_positive_integer_ok(self):
        for pi in ["123", "4567890", "001"]:
            self.assertEqual(int(pi), validatorfuncs.positive_integer(pi))

    @mock.patch("builtins.int")
    def test_positive_integer_raises_ValueError(self, mocked_int):
        mocked_int.return_value = -1
        with self.assertRaises(ValueError):
            validatorfuncs.positive_integer(str(-1))
        for pi in ["", "000", "abc", "-1"]:
            mocked_int.side_effect = ValueError
            with self.assertRaises(ValueError):
                validatorfuncs.positive_integer(pi)

    def test_unsigned_integer_ok(self):
        for ui in ["123", "4567890", "001", "0"]:
            self.assertEqual(int(ui), validatorfuncs.unsigned_integer(ui))

    @mock.patch("builtins.int")
    def test_unsigned_integer_raises_ValueError(self, mocked_int):
        mocked_int.return_value = -1
        with self.assertRaises(ValueError):
            validatorfuncs.unsigned_integer(str(-1))
        for ui in ["", "000", "abc", "-1", "0"]:
            mocked_int.side_effect = ValueError
            with self.assertRaises(ValueError):
                validatorfuncs.unsigned_integer(ui)

    def test_boolean(self):
        for b in ["true", "1", "on", "ENABLED"]:
            self.assertTrue(validatorfuncs.boolean(b))
        for b in ["FalSe", "0", "oFF", "disabled"]:
            self.assertFalse(validatorfuncs.boolean(b))

    def test_boolean_raises_ValueError(self):
        for b in ["", None, 1, 0, True, False, [None], {True: True}]:
            with self.assertRaises(ValueError):
                validatorfuncs.boolean(b)

    def test_timezone_ok(self):
        for tz in ["America/Chicago", "GMT", "UTC"]:
            self.assertEqual(tz, validatorfuncs.timezone(tz).zone)

    def test_timezone_raises_ValueError(self):
        for tz in ["America", None, "", "Mars", "DT"]:
            with self.assertRaises(ValueError):
                validatorfuncs.timezone(tz)

    def test_email_ok(self):
        for e in ["a@a.aa", "zeus@olympus.net"]:
            self.assertEqual(e, validatorfuncs.email(e))

    def test_email_raises_ValueError(self):
        for e in ["", None, ["abc@abc.com"], 123]:
            with self.assertRaises(ValueError):
                validatorfuncs.email(e)

    def test_lock_ok(self):
        for l in ["do:true;look:no", "a:t"]:
            self.assertEqual(l, validatorfuncs.lock(l))

    def test_lock_raises_ValueError(self):
        for l in [";;;", "", ":", ":::", ";:;:", "x:", ":y"]:
            with self.assertRaises(ValueError):
                validatorfuncs.lock(l)
        with self.assertRaises(ValueError):
            validatorfuncs.lock("view:", access_options=())
        with self.assertRaises(ValueError):
            validatorfuncs.lock("view:", access_options=("look"))



"""
Test initial startup procedure

"""

from django.conf import settings
from django.test import TestCase
from mock import MagicMock, patch

from evennia.server import initial_setup


class TestInitialSetup(TestCase):
    @patch("evennia.server.initial_setup.AccountDB")
    def test_get_god_account(self, mocked_accountdb):
        mocked_accountdb.objects.get = MagicMock(return_value=1)
        self.assertEqual(initial_setup._get_superuser_account(), 1)
        mocked_accountdb.objects.get.assert_called_with(id=1)



"""
Pseudo-random generator - vlgeoff 2017

"""

from .random_string_generator import RandomStringGenerator  # noqa
from .random_string_generator import RandomStringGeneratorScript  # noqa
from .random_string_generator import ExhaustedGenerator, RejectedRegex  # noqa



"""
Puzzles System - Provides a typeclass and commands for
objects that can be combined (i.e. 'use'd) to produce
new objects.

Evennia contribution - Henddher 2018

A Puzzle is a recipe of what objects (aka parts) must
be combined by a player so a new set of objects
(aka results) are automatically created.

Installation:

Add the PuzzleSystemCmdSet to all players (e.g. in their Character typeclass).

Alternatively:

    py self.cmdset.add('evennia.contrib.game_systems.puzzles.PuzzleSystemCmdSet')

Usage:

Consider this simple Puzzle:

    orange, mango, yogurt, blender = fruit smoothie

As a Builder:

    create/drop orange
    create/drop mango
    create/drop yogurt
    create/drop blender
    create/drop fruit smoothie

    puzzle smoothie, orange, mango, yogurt, blender = fruit smoothie
    ...
    Puzzle smoothie(#1234) created successfuly.

    destroy/force orange, mango, yogurt, blender, fruit smoothie

    armpuzzle #1234
    Part orange is spawned at ...
    Part mango is spawned at ...
    ....
    Puzzle smoothie(#1234) has been armed successfully

As Player:

    use orange, mango, yogurt, blender
    ...
    Genius, you blended all fruits to create a fruit smoothie!

Details:

Puzzles are created from existing objects. The given
objects are introspected to create prototypes for the
puzzle parts and results. These prototypes become the
puzzle recipe. (See PuzzleRecipe and @puzzle
command). Once the recipe is created, all parts and result
can be disposed (i.e. destroyed).

At a later time, a Builder or a Script can arm the puzzle
and spawn all puzzle parts in their respective
locations (See armpuzzle).

A regular player can collect the puzzle parts and combine
them (See use command). If player has specified
all pieces, the puzzle is considered solved and all
its puzzle parts are destroyed while the puzzle results
are spawened on their corresponding location.


"""

import itertools
from random import choice

from evennia import (
    CmdSet,
    DefaultCharacter,
    DefaultExit,
    DefaultRoom,
    DefaultScript,
    create_script,
)
from evennia.commands.default.muxcommand import MuxCommand
from evennia.prototypes.spawner import spawn
from evennia.utils import logger, search, utils
from evennia.utils.utils import inherits_from

# Tag used by puzzles
_PUZZLES_TAG_CATEGORY = "puzzles"
_PUZZLES_TAG_RECIPE = "puzzle_recipe"
# puzzle part and puzzle result
_PUZZLES_TAG_MEMBER = "puzzle_member"

_PUZZLE_DEFAULT_FAIL_USE_MESSAGE = "You try to utilize %s but nothing happens ... something amiss?"
_PUZZLE_DEFAULT_SUCCESS_USE_MESSAGE = "You are a Genius!!!"
_PUZZLE_DEFAULT_SUCCESS_USE_LOCATION_MESSAGE = "|c{caller}|n performs some kind of tribal dance and |y{result_names}|n seems to appear from thin air"

# ----------- UTILITY FUNCTIONS ------------


def proto_def(obj, with_tags=True):
    """
    Basic properties needed to spawn
    and compare recipe with candidate part
    """
    protodef = {
        # TODO: Don't we need to honor ALL properties? attributes, contents, etc.
        "prototype_key": "%s(%s)" % (obj.key, obj.dbref),
        "key": obj.key,
        "typeclass": obj.typeclass_path,
        "desc": obj.db.desc,
        "location": obj.location,
        "home": obj.home,
        "locks": ";".join(obj.locks.all()),
        "permissions": obj.permissions.all()[:],
    }
    if with_tags:
        tags = obj.tags.all(return_key_and_category=True)
        tags = [(t[0], t[1], None) for t in tags]
        tags.append((_PUZZLES_TAG_MEMBER, _PUZZLES_TAG_CATEGORY, None))
        protodef["tags"] = tags
    return protodef


def maskout_protodef(protodef, mask):
    """
    Returns a new protodef after removing protodef values based on mask
    """
    protodef = dict(protodef)
    for m in mask:
        if m in protodef:
            protodef.pop(m)
    return protodef


# Colorize the default success message
def _colorize_message(msg):
    _i = 0
    _colors = ["|r", "|g", "|y"]
    _msg = []
    for l in msg:
        _msg += _colors[_i] + l
        _i = (_i + 1) % len(_colors)
    msg = "".join(_msg) + "|n"
    return msg


_PUZZLE_DEFAULT_SUCCESS_USE_MESSAGE = _colorize_message(_PUZZLE_DEFAULT_SUCCESS_USE_MESSAGE)

# ------------------------------------------


class PuzzleRecipe(DefaultScript):
    """
    Definition of a Puzzle Recipe
    """

    def save_recipe(self, puzzle_name, parts, results):
        self.db.puzzle_name = str(puzzle_name)
        self.db.parts = tuple(parts)
        self.db.results = tuple(results)
        self.db.mask = tuple()
        self.tags.add(_PUZZLES_TAG_RECIPE, category=_PUZZLES_TAG_CATEGORY)
        self.db.use_success_message = _PUZZLE_DEFAULT_SUCCESS_USE_MESSAGE
        self.db.use_success_location_message = _PUZZLE_DEFAULT_SUCCESS_USE_LOCATION_MESSAGE


class CmdCreatePuzzleRecipe(MuxCommand):
    """
    Creates a puzzle recipe. A puzzle consists of puzzle-parts that
    the player can 'use' together to create a specified result.

    Usage:
        @puzzle name,<part1[,part2,...>] = <result1[,result2,...]>

    Example:
        create/drop balloon
        create/drop glass of water
        create/drop water balloon
        @puzzle waterballon,balloon,glass of water = water balloon
        @del ballon, glass of water, water balloon
        @armpuzzle #1

    Notes:
    Each part and result are objects that must (temporarily) exist and be placed in their
    corresponding location in order to create the puzzle. After the creation of the puzzle,
    these objects are not needed anymore and can be deleted. Components of the puzzle
    will be re-created by use of the `@armpuzzle` command later.

    """

    key = "@puzzle"
    aliases = "@puzzlerecipe"
    locks = "cmd:perm(puzzle) or perm(Builder)"
    help_category = "Puzzles"

    confirm = True
    default_confirm = "no"

    def func(self):
        caller = self.caller

        if len(self.lhslist) < 2 or not self.rhs:
            string = "Usage: @puzzle name,<part1[,...]> = <result1[,...]>"
            caller.msg(string)
            return

        puzzle_name = self.lhslist[0]
        if len(puzzle_name) == 0:
            caller.msg("Invalid puzzle name %r." % puzzle_name)
            return

        # if there is another puzzle with same name
        # warn user that parts and results will be
        # interchangable
        _puzzles = search.search_script_attribute(key="puzzle_name", value=puzzle_name)
        _puzzles = list(filter(lambda p: isinstance(p, PuzzleRecipe), _puzzles))
        if _puzzles:
            confirm = (
                "There are %d puzzles with the same name.\n" % len(_puzzles)
                + "Its parts and results will be interchangeable.\n"
                + "Continue yes/[no]? "
            )
            answer = ""
            while answer.strip().lower() not in ("y", "yes", "n", "no"):
                answer = yield (confirm)
                answer = self.default_confirm if answer == "" else answer
            if answer.strip().lower() in ("n", "no"):
                caller.msg("Cancelled: no puzzle created.")
                return

        def is_valid_obj_location(obj):
            valid = True
            # Rooms are the only valid locations.
            # TODO: other valid locations could be added here.
            # Certain locations can be handled accordingly: e.g,
            # a part is located in a character's inventory,
            # perhaps will translate into the player character
            # having the part in his/her inventory while being
            # located in the same room where the builder was
            # located.
            # Parts and results may have different valid locations
            if not inherits_from(obj.location, DefaultRoom):
                caller.msg("Invalid location for %s" % (obj.key))
                valid = False
            return valid

        def is_valid_part_location(part):
            return is_valid_obj_location(part)

        def is_valid_result_location(part):
            return is_valid_obj_location(part)

        def is_valid_inheritance(obj):
            valid = (
                not inherits_from(obj, DefaultCharacter)
                and not inherits_from(obj, DefaultRoom)
                and not inherits_from(obj, DefaultExit)
            )
            if not valid:
                caller.msg("Invalid typeclass for %s" % (obj))
            return valid

        def is_valid_part(part):
            return is_valid_inheritance(part) and is_valid_part_location(part)

        def is_valid_result(result):
            return is_valid_inheritance(result) and is_valid_result_location(result)

        parts = []
        for objname in self.lhslist[1:]:
            obj = caller.search(objname)
            if not obj:
                return
            if not is_valid_part(obj):
                return
            parts.append(obj)

        results = []
        for objname in self.rhslist:
            obj = caller.search(objname)
            if not obj:
                return
            if not is_valid_result(obj):
                return
            results.append(obj)

        for part in parts:
            caller.msg("Part %s(%s)" % (part.name, part.dbref))

        for result in results:
            caller.msg("Result %s(%s)" % (result.name, result.dbref))

        proto_parts = [proto_def(obj) for obj in parts]
        proto_results = [proto_def(obj) for obj in results]

        puzzle = create_script(PuzzleRecipe, key=puzzle_name, persistent=True)
        puzzle.save_recipe(puzzle_name, proto_parts, proto_results)
        puzzle.locks.add("control:id(%s) or perm(Builder)" % caller.dbref[1:])

        caller.msg(
            "Puzzle |y'%s' |w%s(%s)|n has been created |gsuccessfully|n."
            % (puzzle.db.puzzle_name, puzzle.name, puzzle.dbref)
        )

        caller.msg(
            "You may now dispose of all parts and results. \n"
            "Use @puzzleedit #{dbref} to customize this puzzle further. \n"
            "Use @armpuzzle #{dbref} to arm a new puzzle instance.".format(dbref=puzzle.dbref)
        )


class CmdEditPuzzle(MuxCommand):
    """
    Edits puzzle properties

    Usage:
        @puzzleedit[/delete] <#dbref>
        @puzzleedit <#dbref>/use_success_message = <Custom message>
        @puzzleedit <#dbref>/use_success_location_message = <Custom message from {caller} producing {result_names}>
        @puzzleedit <#dbref>/mask = attr1[,attr2,...]>
        @puzzleedit[/addpart] <#dbref> = <obj[,obj2,...]>
        @puzzleedit[/delpart] <#dbref> = <obj[,obj2,...]>
        @puzzleedit[/addresult] <#dbref> = <obj[,obj2,...]>
        @puzzleedit[/delresult] <#dbref> = <obj[,obj2,...]>

    Switches:
      addpart - adds parts to the puzzle
      delpart - removes parts from the puzzle
      addresult - adds results to the puzzle
      delresult - removes results from the puzzle
      delete - deletes the recipe. Existing parts and results aren't modified

      mask - attributes to exclude during matching (e.g. location, desc, etc.)
      use_success_location_message containing {result_names} and {caller} will
        automatically be replaced with correct values. Both are optional.

      When removing parts/results, it's possible to remove all.

    """

    key = "@puzzleedit"
    locks = "cmd:perm(puzzleedit) or perm(Builder)"
    help_category = "Puzzles"

    def func(self):
        self._USAGE = "Usage: @puzzleedit[/switches] <dbref>[/attribute = <value>]"
        caller = self.caller

        if not self.lhslist:
            caller.msg(self._USAGE)
            return

        if "/" in self.lhslist[0]:
            recipe_dbref, attr = self.lhslist[0].split("/")
        else:
            recipe_dbref = self.lhslist[0]

        if not utils.dbref(recipe_dbref):
            caller.msg("A puzzle recipe's #dbref must be specified.\n" + self._USAGE)
            return

        puzzle = search.search_script(recipe_dbref)
        if not puzzle or not inherits_from(puzzle[0], PuzzleRecipe):
            caller.msg("%s(%s) is not a puzzle" % (puzzle[0].name, recipe_dbref))
            return

        puzzle = puzzle[0]
        puzzle_name_id = "%s(%s)" % (puzzle.name, puzzle.dbref)

        if "delete" in self.switches:
            if not (puzzle.access(caller, "control") or puzzle.access(caller, "delete")):
                caller.msg("You don't have permission to delete %s." % puzzle_name_id)
                return

            puzzle.delete()
            caller.msg("%s was deleted" % puzzle_name_id)
            return

        elif "addpart" in self.switches:
            objs = self._get_objs()
            if objs:
                added = self._add_parts(objs, puzzle)
                caller.msg("%s were added to parts" % (", ".join(added)))
            return

        elif "delpart" in self.switches:
            objs = self._get_objs()
            if objs:
                removed = self._remove_parts(objs, puzzle)
                caller.msg("%s were removed from parts" % (", ".join(removed)))
            return

        elif "addresult" in self.switches:
            objs = self._get_objs()
            if objs:
                added = self._add_results(objs, puzzle)
                caller.msg("%s were added to results" % (", ".join(added)))
            return

        elif "delresult" in self.switches:
            objs = self._get_objs()
            if objs:
                removed = self._remove_results(objs, puzzle)
                caller.msg("%s were removed from results" % (", ".join(removed)))
            return

        else:
            # edit attributes

            if not (puzzle.access(caller, "control") or puzzle.access(caller, "edit")):
                caller.msg("You don't have permission to edit %s." % puzzle_name_id)
                return

            if attr == "use_success_message":
                puzzle.db.use_success_message = self.rhs
                caller.msg(
                    "%s use_success_message = %s\n"
                    % (puzzle_name_id, puzzle.db.use_success_message)
                )
                return
            elif attr == "use_success_location_message":
                puzzle.db.use_success_location_message = self.rhs
                caller.msg(
                    "%s use_success_location_message = %s\n"
                    % (puzzle_name_id, puzzle.db.use_success_location_message)
                )
                return
            elif attr == "mask":
                puzzle.db.mask = tuple(self.rhslist)
                caller.msg("%s mask = %r\n" % (puzzle_name_id, puzzle.db.mask))
                return

    def _get_objs(self):
        if not self.rhslist:
            self.caller.msg(self._USAGE)
            return
        objs = []
        for o in self.rhslist:
            obj = self.caller.search(o)
            if obj:
                objs.append(obj)
        return objs

    def _add_objs_to(self, objs, to):
        """Adds propto objs to the given set (parts or results)"""
        added = []
        toobjs = list(to[:])
        for obj in objs:
            protoobj = proto_def(obj)
            toobjs.append(protoobj)
            added.append(obj.key)
        return added, toobjs

    def _remove_objs_from(self, objs, frm):
        """Removes propto objs from the given set (parts or results)"""
        removed = []
        fromobjs = list(frm[:])
        for obj in objs:
            protoobj = proto_def(obj)
            if protoobj in fromobjs:
                fromobjs.remove(protoobj)
                removed.append(obj.key)
        return removed, fromobjs

    def _add_parts(self, objs, puzzle):
        added, toobjs = self._add_objs_to(objs, puzzle.db.parts)
        puzzle.db.parts = tuple(toobjs)
        return added

    def _remove_parts(self, objs, puzzle):
        removed, fromobjs = self._remove_objs_from(objs, puzzle.db.parts)
        puzzle.db.parts = tuple(fromobjs)
        return removed

    def _add_results(self, objs, puzzle):
        added, toobjs = self._add_objs_to(objs, puzzle.db.results)
        puzzle.db.results = tuple(toobjs)
        return added

    def _remove_results(self, objs, puzzle):
        removed, fromobjs = self._remove_objs_from(objs, puzzle.db.results)
        puzzle.db.results = tuple(fromobjs)
        return removed


class CmdArmPuzzle(MuxCommand):
    """
    Arms a puzzle by spawning all its parts.

    Usage:
      @armpuzzle <puzzle #dbref>

    Notes:
        Create puzzles with `@puzzle`; get list of
        defined puzzles using `@lspuzzlerecipes`.

    """

    key = "@armpuzzle"
    locks = "cmd:perm(armpuzzle) or perm(Builder)"
    help_category = "Puzzles"

    def func(self):
        caller = self.caller

        if self.args is None or not utils.dbref(self.args):
            caller.msg("A puzzle recipe's #dbref must be specified")
            return

        puzzle = search.search_script(self.args)
        if not puzzle or not inherits_from(puzzle[0], PuzzleRecipe):
            caller.msg("Invalid puzzle %r" % (self.args))
            return

        puzzle = puzzle[0]
        caller.msg(
            "Puzzle Recipe %s(%s) '%s' found.\nSpawning %d parts ..."
            % (puzzle.name, puzzle.dbref, puzzle.db.puzzle_name, len(puzzle.db.parts))
        )

        for proto_part in puzzle.db.parts:
            part = spawn(proto_part)[0]
            caller.msg(
                "Part %s(%s) spawned and placed at %s(%s)"
                % (part.name, part.dbref, part.location, part.location.dbref)
            )
            part.tags.add(puzzle.db.puzzle_name, category=_PUZZLES_TAG_CATEGORY)
            part.db.puzzle_name = puzzle.db.puzzle_name

        caller.msg("Puzzle armed |gsuccessfully|n.")


def _lookups_parts_puzzlenames_protodefs(parts):
    # Create lookup dicts by part's dbref and by puzzle_name(tags)
    parts_dict = dict()
    puzzlename_tags_dict = dict()
    puzzle_ingredients = dict()
    for part in parts:
        parts_dict[part.dbref] = part
        protodef = proto_def(part, with_tags=False)
        # remove 'prototype_key' as it will prevent equality
        del protodef["prototype_key"]
        puzzle_ingredients[part.dbref] = protodef
        tags_categories = part.tags.all(return_key_and_category=True)
        for tag, category in tags_categories:
            if category != _PUZZLES_TAG_CATEGORY:
                continue
            if tag not in puzzlename_tags_dict:
                puzzlename_tags_dict[tag] = []
            puzzlename_tags_dict[tag].append(part.dbref)
    return parts_dict, puzzlename_tags_dict, puzzle_ingredients


def _puzzles_by_names(names):
    # Find all puzzles by puzzle name (i.e. tag name)
    puzzles = []
    for puzzle_name in names:
        _puzzles = search.search_script_attribute(key="puzzle_name", value=puzzle_name)
        _puzzles = list(filter(lambda p: isinstance(p, PuzzleRecipe), _puzzles))
        if not _puzzles:
            continue
        else:
            puzzles.extend(_puzzles)
    return puzzles


def _matching_puzzles(puzzles, puzzlename_tags_dict, puzzle_ingredients):
    # Check if parts can be combined to solve a puzzle
    matched_puzzles = dict()
    for puzzle in puzzles:
        puzzle_protoparts = list(puzzle.db.parts[:])
        puzzle_mask = puzzle.db.mask[:]
        # remove tags and prototype_key as they prevent equality
        for i, puzzle_protopart in enumerate(puzzle_protoparts[:]):
            del puzzle_protopart["tags"]
            del puzzle_protopart["prototype_key"]
            puzzle_protopart = maskout_protodef(puzzle_protopart, puzzle_mask)
            puzzle_protoparts[i] = puzzle_protopart

        matched_dbrefparts = []
        parts_dbrefs = puzzlename_tags_dict[puzzle.db.puzzle_name]
        for part_dbref in parts_dbrefs:
            protopart = puzzle_ingredients[part_dbref]
            protopart = maskout_protodef(protopart, puzzle_mask)
            if protopart in puzzle_protoparts:
                puzzle_protoparts.remove(protopart)
                matched_dbrefparts.append(part_dbref)
        else:
            if len(puzzle_protoparts) == 0:
                matched_puzzles[puzzle.dbref] = matched_dbrefparts
    return matched_puzzles


class CmdUsePuzzleParts(MuxCommand):
    """
    Use an object, or a group of objects at once.


    Example:
      You look around you and see a pole, a long string, and a needle.

      use pole, long string, needle

      Genius! You built a fishing pole.


    Usage:
        use <obj1> [,obj2,...]
    """

    # Technical explanation
    """
    Searches for all puzzles whose parts match the given set of objects. If there are matching
    puzzles, the result objects are spawned in their corresponding location if all parts have been
    passed in.
    """

    key = "use"
    aliases = "combine"
    locks = "cmd:pperm(use) or pperm(Player)"
    help_category = "Puzzles"

    def func(self):
        caller = self.caller

        if not self.lhs:
            caller.msg("Use what?")
            return

        many = "these" if len(self.lhslist) > 1 else "this"

        # either all are parts, or abort finding matching puzzles
        parts = []
        partnames = self.lhslist[:]
        for partname in partnames:
            part = caller.search(
                partname,
                multimatch_string="Which %s. There are many.\n" % (partname),
                nofound_string="There is no %s around." % (partname),
            )

            if not part:
                return

            if not part.tags.get(_PUZZLES_TAG_MEMBER, category=_PUZZLES_TAG_CATEGORY):
                # not a puzzle part ... abort
                caller.msg("You have no idea how %s can be used" % (many))
                return

            # a valid part
            parts.append(part)

        # Create lookup dicts by part's dbref and by puzzle_name(tags)
        parts_dict, puzzlename_tags_dict, puzzle_ingredients = _lookups_parts_puzzlenames_protodefs(
            parts
        )

        # Find all puzzles by puzzle name (i.e. tag name)
        puzzles = _puzzles_by_names(puzzlename_tags_dict.keys())

        logger.log_info("PUZZLES %r" % ([(p.dbref, p.db.puzzle_name) for p in puzzles]))

        # Create lookup dict of puzzles by dbref
        puzzles_dict = dict((puzzle.dbref, puzzle) for puzzle in puzzles)
        # Check if parts can be combined to solve a puzzle
        matched_puzzles = _matching_puzzles(puzzles, puzzlename_tags_dict, puzzle_ingredients)

        if len(matched_puzzles) == 0:
            # TODO: we could use part.fail_message instead, if there was one
            #    random part falls and lands on your feet
            #    random part hits you square on the face
            caller.msg(_PUZZLE_DEFAULT_FAIL_USE_MESSAGE % (many))
            return

        puzzletuples = sorted(matched_puzzles.items(), key=lambda t: len(t[1]), reverse=True)

        logger.log_info("MATCHED PUZZLES %r" % (puzzletuples))

        # sort all matched puzzles and pick largest one(s)
        puzzledbref, matched_dbrefparts = puzzletuples[0]
        nparts = len(matched_dbrefparts)
        puzzle = puzzles_dict[puzzledbref]
        largest_puzzles = list(itertools.takewhile(lambda t: len(t[1]) == nparts, puzzletuples))

        # if there are more than one, choose one at random.
        # we could show the names of all those that can be resolved
        # but that would give away that there are other puzzles that
        # can be resolved with the same parts.
        # just hint how many.
        if len(largest_puzzles) > 1:
            caller.msg(
                "Your gears start turning and %d different ideas come to your mind ...\n"
                % (len(largest_puzzles))
            )
            puzzletuple = choice(largest_puzzles)
            puzzle = puzzles_dict[puzzletuple[0]]
            caller.msg("You try %s ..." % (puzzle.db.puzzle_name))

        # got one, spawn its results
        result_names = []
        for proto_result in puzzle.db.results:
            result = spawn(proto_result)[0]
            result.tags.add(puzzle.db.puzzle_name, category=_PUZZLES_TAG_CATEGORY)
            result.db.puzzle_name = puzzle.db.puzzle_name
            result_names.append(result.name)

        # Destroy all parts used
        for dbref in matched_dbrefparts:
            parts_dict[dbref].delete()

        result_names = ", ".join(result_names)
        caller.msg(puzzle.db.use_success_message)
        caller.location.msg_contents(
            puzzle.db.use_success_location_message.format(caller=caller, result_names=result_names),
            exclude=(caller,),
        )


class CmdListPuzzleRecipes(MuxCommand):
    """
    Searches for all puzzle recipes

    Usage:
        @lspuzzlerecipes
    """

    key = "@lspuzzlerecipes"
    locks = "cmd:perm(lspuzzlerecipes) or perm(Builder)"
    help_category = "Puzzles"

    def func(self):
        caller = self.caller

        recipes = search.search_script_tag(_PUZZLES_TAG_RECIPE, category=_PUZZLES_TAG_CATEGORY)

        div = "-" * 60
        text = [div]
        msgf_recipe = "Puzzle |y'%s' %s(%s)|n"
        msgf_item = "%2s|c%15s|n: |w%s|n"
        for recipe in recipes:
            text.append(msgf_recipe % (recipe.db.puzzle_name, recipe.name, recipe.dbref))
            text.append("Success Caller message:\n" + recipe.db.use_success_message + "\n")
            text.append(
                "Success Location message:\n" + recipe.db.use_success_location_message + "\n"
            )
            text.append("Mask:\n" + str(recipe.db.mask) + "\n")
            text.append("Parts")
            for protopart in recipe.db.parts[:]:
                mark = "-"
                for k, v in protopart.items():
                    text.append(msgf_item % (mark, k, v))
                    mark = ""
            text.append("Results")
            for protoresult in recipe.db.results[:]:
                mark = "-"
                for k, v in protoresult.items():
                    text.append(msgf_item % (mark, k, v))
                    mark = ""
        else:
            text.append(div)
            text.append("Found |r%d|n puzzle(s)." % (len(recipes)))
            text.append(div)
        caller.msg("\n".join(text))


class CmdListArmedPuzzles(MuxCommand):
    """
    Searches for all armed puzzles

    Usage:
        @lsarmedpuzzles
    """

    key = "@lsarmedpuzzles"
    locks = "cmd:perm(lsarmedpuzzles) or perm(Builder)"
    help_category = "Puzzles"

    def func(self):
        caller = self.caller

        armed_puzzles = search.search_tag(_PUZZLES_TAG_MEMBER, category=_PUZZLES_TAG_CATEGORY)

        armed_puzzles = dict(
            (k, list(g)) for k, g in itertools.groupby(armed_puzzles, lambda ap: ap.db.puzzle_name)
        )

        div = "-" * 60
        msgf_pznm = "Puzzle name: |y%s|n"
        msgf_item = "|m%25s|w(%s)|n at |c%25s|w(%s)|n"
        text = [div]
        for pzname, items in armed_puzzles.items():
            text.append(msgf_pznm % (pzname))
            for item in items:
                text.append(
                    msgf_item % (item.name, item.dbref, item.location.name, item.location.dbref)
                )
        else:
            text.append(div)
            text.append("Found |r%d|n armed puzzle(s)." % (len(armed_puzzles)))
            text.append(div)
        caller.msg("\n".join(text))


class PuzzleSystemCmdSet(CmdSet):
    """
    CmdSet to create, arm and resolve Puzzles
    """

    def at_cmdset_creation(self):
        super().at_cmdset_creation()

        self.add(CmdCreatePuzzleRecipe())
        self.add(CmdEditPuzzle())
        self.add(CmdArmPuzzle())
        self.add(CmdListPuzzleRecipes())
        self.add(CmdListArmedPuzzles())
        self.add(CmdUsePuzzleParts())



"""
Pseudo-random generator and registry

Evennia contribution - Vincent Le Goff 2017

This contrib can be used to generate pseudo-random strings of information
with specific criteria.  You could, for instance, use it to generate
phone numbers, license plate numbers, validation codes, non-sensivite
passwords and so on.  The strings generated by the generator will be
stored and won't be available again in order to avoid repetition.
Here's a very simple example:

```python
from evennia.contrib.utils.random_string_generator import RandomStringGenerator
# Create a generator for phone numbers
phone_generator = RandomStringGenerator("phone number", r"555-[0-9]{3}-[0-9]{4}")
# Generate a phone number (555-XXX-XXXX with X as numbers)
number = phone_generator.get()
# `number` will contain something like: "555-981-2207"
# If you call `phone_generator.get`, it won't give the same anymore.phone_generator.all()
# Will return a list of all currently-used phone numbers
phone_generator.remove("555-981-2207")
# The number can be generated again
```

To use it, you will need to:

1. Import the `RandomStringGenerator` class from the contrib.
2. Create an instance of this class taking two arguments:
   - The name of the gemerator (like "phone number", "license plate"...).
   - The regular expression representing the expected results.
3. Use the generator's `all`, `get` and `remove` methods as shown above.

To understand how to read and create regular expressions, you can refer to
[the documentation on the re module](https://docs.python.org/2/library/re.html).
Some examples of regular expressions you could use:

- `r"555-\d{3}-\d{4}"`: 555, a dash, 3 digits, another dash, 4 digits.
- `r"[0-9]{3}[A-Z][0-9]{3}"`: 3 digits, a capital letter, 3 digits.
- `r"[A-Za-z0-9]{8,15}"`: between 8 and 15 letters and digits.
- ...

Behind the scenes, a script is created to store the generated information
for a single generator.  The `RandomStringGenerator` object will also
read the regular expression you give to it to see what information is
required (letters, digits, a more restricted class, simple characters...)...
More complex regular expressions (with branches for instance) might not be
available.

"""

import re
import string
import time
from random import choice, randint, seed

from evennia import DefaultScript, ScriptDB
from evennia.utils.create import create_script


class RejectedRegex(RuntimeError):

    """The provided regular expression has been rejected.

    More details regarding why this error occurred will be provided in
    the message.  The usual reason is the provided regular expression is
    not specific enough and could lead to inconsistent generating.

    """

    pass


class ExhaustedGenerator(RuntimeError):

    """The generator hasn't any available strings to generate anymore."""

    pass


class RandomStringGeneratorScript(DefaultScript):

    """
    The global script to hold all generators.

    It will be automatically created the first time `generate` is called
    on a RandomStringGenerator object.

    """

    def at_script_creation(self):
        """Hook called when the script is created."""
        self.key = "generator_script"
        self.desc = "Global generator script"
        self.persistent = True

        # Permanent data to be stored
        self.db.generated = {}


class RandomStringGenerator:

    """
    A generator class to generate pseudo-random strings with a rule.

    The "rule" defining what the generator should provide in terms of
    string is given as a regular expression when creating instances of
    this class.  You can use the `all` method to get all generated strings,
    the `get` method to generate a new string, the `remove` method
    to remove a generated string, or the `clear` method to remove all
    generated strings.

    Bear in mind, however, that while the generated strings will be
    stored to avoid repetition, the generator will not concern itself
    with how the string is stored on the object you use.  You probably
    want to create a tag to mark this object.  This is outside of the scope
    of this class.

    """

    # We keep the script as a class variable to optimize querying
    # with multiple instandces
    script = None

    def __init__(self, name, regex):
        """
        Create a new generator.

        Args:
            name (str): name of the generator to create.
            regex (str): regular expression describing the generator.

        Notes:
            `name` should be an explicit name.  If you use more than one
            generator in your game, be sure to give them different names.
            This name will be used to store the generated information
            in the global script, and in case of errors.

            The regular expression should describe the generator, what
            it should generate: a phone number, a license plate, a password
            or something else.  Regular expressions allow you to use
            pretty advanced criteria, but be aware that some regular
            expressions will be rejected if not specific enough.

        Raises:
            RejectedRegex: the provided regular expression couldn't be
            accepted as a valid generator description.

        """
        self.name = name
        self.elements = []
        self.total = 1

        # Analyze the regex if any
        if regex:
            self._find_elements(regex)

    def __repr__(self):
        return (
            "<evennia.contrib.utils.random_string_generator.RandomStringGenerator for {}>".format(
                self.name
            )
        )

    def _get_script(self):
        """Get or create the script."""
        if type(self).script:
            return type(self).script

        try:
            script = ScriptDB.objects.get(db_key="generator_script")
        except ScriptDB.DoesNotExist:
            script = create_script(
                "evennia.contrib.utils.random_string_generator.RandomStringGeneratorScript"
            )

        type(self).script = script
        return script

    def _find_elements(self, regex):
        """
        Find the elements described in the regular expression.  This will
        analyze the provided regular expression and try to find elements.

        Args:
            regex (str): the regular expression.

        """
        try:
            # python 3.11
            regex_parser = re._parser
        except AttributeError:
            # python <3.11
            regex_parser = re.sre_parse

        self.total = 1
        self.elements = []
        tree = regex_parser.parse(regex).data  # note - sre_parse removed in py3.11
        # `tree` contains a list of elements in the regular expression
        for element in tree:
            # `element` is also a list, the first element is a string
            name = str(element[0]).lower()
            desc = {"min": 1, "max": 1}

            # If `.`, break here
            if name == "any":
                raise RejectedRegex(
                    "the . definition is too broad, specify what you need more precisely"
                )
            elif name == "at":
                # Either the beginning or end, we ignore it
                continue
            elif name == "min_repeat":
                raise RejectedRegex("you have to provide a maximum number of this character class")
            elif name == "max_repeat":
                desc["min"] = element[1][0]
                desc["max"] = element[1][1]
                desc["chars"] = self._find_literal(element[1][2][0])
            elif name == "in":
                desc["chars"] = self._find_literal(element)
            elif name == "literal":
                desc["chars"] = self._find_literal(element)
            else:
                raise RejectedRegex("unhandled regex syntax:: {}".format(repr(name)))

            self.elements.append(desc)
            self.total *= len(desc["chars"]) ** desc["max"]

    def _find_literal(self, element):
        """Find the literal corresponding to a piece of regular expression."""
        name = str(element[0]).lower()
        chars = []
        if name == "literal":
            chars.append(chr(element[1]))
        elif name == "in":
            negate = False
            if element[1][0][0] == "negate":
                negate = True
                chars = list(string.ascii_letters + string.digits)

            for part in element[1]:
                if part[0] == "negate":
                    continue

                sublist = self._find_literal(part)
                for char in sublist:
                    if negate:
                        if char in chars:
                            chars.remove(char)
                    else:
                        chars.append(char)
        elif name == "range":
            chars = [chr(i) for i in range(element[1][0], element[1][1] + 1)]
        elif name == "category":
            category = str(element[1]).lower()
            if category == "category_digit":
                chars = list(string.digits)
            elif category == "category_word":
                chars = list(string.letters)
            else:
                raise RejectedRegex("unknown category: {}".format(category))
        else:
            raise RejectedRegex("cannot find the literal: {}".format(element[0]))

        return chars

    def all(self):
        """
        Return all generated strings for this generator.

        Returns:
            strings (list of strr): the list of strings that are already
            used.  The strings that were generated first come first in the list.

        """
        script = self._get_script()
        generated = list(script.db.generated.get(self.name, []))
        return generated

    def get(self, store=True, unique=True):
        """
        Generate a pseudo-random string according to the regular expression.

        Args:
            store (bool, optional): store the generated string in the script.
            unique (bool, optional): keep on trying if the string is already used.

        Returns:
            The newly-generated string.

        Raises:
            ExhaustedGenerator: if there's no available string in this generator.

        Note:
            Unless asked explicitly, the returned string can't repeat itself.

        """
        script = self._get_script()
        generated = script.db.generated.get(self.name)
        if generated is None:
            script.db.generated[self.name] = []
            generated = script.db.generated[self.name]

        if len(generated) >= self.total:
            raise ExhaustedGenerator

        # Generate a pseudo-random string that might be used already
        result = ""
        for element in self.elements:
            number = randint(element["min"], element["max"])
            chars = element["chars"]
            for index in range(number):
                char = choice(chars)
                result += char

        # If the string has already been generated, try again
        if result in generated and unique:
            # Change the random seed, incrementing it slowly
            epoch = time.time()
            while result in generated:
                epoch += 1
                seed(epoch)
                result = self.get(store=False, unique=False)

        if store:
            generated.append(result)

        return result

    def remove(self, element):
        """
        Remove a generated string from the list of stored strings.

        Args:
            element (str): the string to remove from the list of generated strings.

        Raises:
            ValueError: the specified value hasn't been generated and is not present.

        Note:
            The specified string has to be present in the script (so
            has to have been generated).  It will remove this entry
            from the script, so this string could be generated again by
            calling the `get` method.

        """
        script = self._get_script()
        generated = script.db.generated.get(self.name, [])
        if element not in generated:
            raise ValueError(
                "the string {} isn't stored as generated by the generator {}".format(
                    element, self.name
                )
            )

        generated.remove(element)

    def clear(self):
        """
        Clear the generator of all generated strings.

        """
        script = self._get_script()
        generated = script.db.generated.get(self.name, [])
        generated[:] = []



"""
The Evennia Portal service acts as an AMP-server, handling AMP
communication to the AMP clients connecting to it (by default
these are the Evennia Server and the evennia launcher).

"""
import os
import sys
from subprocess import STDOUT, Popen

from django.conf import settings
from twisted.internet import protocol

import evennia
from evennia.server.portal import amp
from evennia.utils import logger
from evennia.utils.utils import class_from_module


def _is_windows():
    return os.name == "nt"


def getenv():
    """
    Get current environment and add PYTHONPATH.

    Returns:
        env (dict): Environment global dict.

    """
    sep = ";" if _is_windows() else ":"
    env = os.environ.copy()
    env["PYTHONPATH"] = sep.join(sys.path)
    return env


class AMPServerFactory(protocol.ServerFactory):

    """
    This factory creates AMP Server connection. This acts as the 'Portal'-side communication to the
    'Server' process.

    """

    noisy = False

    def logPrefix(self):
        """
        How this is named in logs

        """
        return "AMP"

    def __init__(self, portal):
        """
        Initialize the factory. This is called as the Portal service starts.

        Args:
            portal (Portal): The Evennia Portal service instance.
            protocol (Protocol): The protocol the factory creates
                instances of.

        """
        self.portal = portal
        self.protocol = class_from_module(settings.AMP_SERVER_PROTOCOL_CLASS)
        self.broadcasts = []
        self.server_connection = None
        self.launcher_connection = None
        self.disconnect_callbacks = {}
        self.server_connect_callbacks = []

    def buildProtocol(self, addr):
        """
        Start a new connection, and store it on the service object.

        Args:
            addr (str): Connection address. Not used.

        Returns:
            protocol (Protocol): The created protocol.

        """
        self.portal.amp_protocol = self.protocol()
        self.portal.amp_protocol.factory = self
        return self.portal.amp_protocol


class AMPServerProtocol(amp.AMPMultiConnectionProtocol):
    """
    Protocol subclass for the AMP-server run by the Portal.

    """

    def connectionLost(self, reason):
        """
        Set up a simple callback mechanism to let the amp-server wait for a connection to close.

        """
        # wipe broadcast and data memory
        super().connectionLost(reason)
        if self.factory.server_connection == self:
            self.factory.server_connection = None
            self.factory.portal.server_info_dict = {}
        if self.factory.launcher_connection == self:
            self.factory.launcher_connection = None

        callback, args, kwargs = self.factory.disconnect_callbacks.pop(self, (None, None, None))
        if callback:
            try:
                callback(*args, **kwargs)
            except Exception:
                logger.log_trace()

    def get_status(self):
        """
        Return status for the Evennia infrastructure.

        Returns:
            status (tuple): The portal/server status and pids
                (portal_live, server_live, portal_PID, server_PID).

        """
        server_connected = bool(
            self.factory.server_connection and self.factory.server_connection.transport.connected
        )
        portal_info_dict = self.factory.portal.get_info_dict()
        server_info_dict = self.factory.portal.server_info_dict
        server_pid = self.factory.portal.server_process_id
        portal_pid = os.getpid()
        return (True, server_connected, portal_pid, server_pid, portal_info_dict, server_info_dict)

    def data_to_server(self, command, sessid, **kwargs):
        """
        Send data across the wire to the Server.

        Args:
            command (AMP Command): A protocol send command.
            sessid (int): A unique Session id.
            kwargs (any): Data to send. This will be pickled.

        Returns:
            deferred (deferred or None): A deferred with an errback.

        Notes:
            Data will be sent across the wire pickled as a tuple
            (sessid, kwargs).

        """
        # print("portal data_to_server: {}, {}, {}".format(command, sessid, kwargs))
        if self.factory.server_connection:
            return self.factory.server_connection.callRemote(
                command, packed_data=amp.dumps((sessid, kwargs))
            ).addErrback(self.errback, command.key)
        else:
            # if no server connection is available, broadcast
            return self.broadcast(command, sessid, packed_data=amp.dumps((sessid, kwargs)))

    def start_server(self, server_twistd_cmd):
        """
        (Re-)Launch the Evennia server.

        Args:
            server_twisted_cmd (list): The server start instruction
                to pass to POpen to start the server.

        """
        # start the Server
        print("Portal starting server ... ")
        process = None
        with open(settings.SERVER_LOG_FILE, "a") as logfile:
            # we link stdout to a file in order to catch
            # eventual errors happening before the Server has
            # opened its logger.
            try:
                if _is_windows():
                    # Windows requires special care
                    create_no_window = 0x08000000
                    process = Popen(
                        server_twistd_cmd,
                        env=getenv(),
                        bufsize=-1,
                        stdout=logfile,
                        stderr=STDOUT,
                        creationflags=create_no_window,
                    )

                else:
                    process = Popen(
                        server_twistd_cmd, env=getenv(), bufsize=-1, stdout=logfile, stderr=STDOUT
                    )
            except Exception:
                logger.log_trace()

            self.factory.portal.server_twistd_cmd = server_twistd_cmd
            logfile.flush()
        if process and not _is_windows():
            # avoid zombie-process on Unix/BSD
            process.wait()
        # unset the reset-mode flag on the portal
        self.factory.portal.server_restart_mode = None
        return

    def wait_for_disconnect(self, callback, *args, **kwargs):
        """
        Add a callback for when this connection is lost.

        Args:
            callback (callable): Will be called with *args, **kwargs
                once this protocol is disconnected.

        """
        self.factory.disconnect_callbacks[self] = (callback, args, kwargs)

    def wait_for_server_connect(self, callback, *args, **kwargs):
        """
        Add a callback for when the Server is sure to have connected.

        Args:
            callback (callable): Will be called with *args, **kwargs
                once the Server handshake with Portal is complete.

        """
        self.factory.server_connect_callbacks.append((callback, args, kwargs))

    def stop_server(self, mode="shutdown"):
        """
        Shut down server in one or more modes.

        Args:
            mode (str): One of 'shutdown', 'reload' or 'reset'.

        """
        if mode == "reload":
            self.send_AdminPortal2Server(amp.DUMMYSESSION, operation=amp.SRELOAD)
        elif mode == "reset":
            self.send_AdminPortal2Server(amp.DUMMYSESSION, operation=amp.SRESET)
        elif mode == "shutdown":
            self.send_AdminPortal2Server(amp.DUMMYSESSION, operation=amp.SSHUTD)
        self.factory.portal.server_restart_mode = mode

    # sending amp data

    def send_Status2Launcher(self):
        """
        Send a status stanza to the launcher.

        """
        # print("send status to launcher")
        # print("self.get_status(): {}".format(self.get_status()))
        if self.factory.launcher_connection:
            self.factory.launcher_connection.callRemote(
                amp.MsgStatus, status=amp.dumps(self.get_status())
            ).addErrback(self.errback, amp.MsgStatus.key)

    def send_MsgPortal2Server(self, session, **kwargs):
        """
        Access method called by the Portal and executed on the Portal.

        Args:
            session (session): Session
            kwargs (any, optional): Optional data.

        Returns:
            deferred (Deferred): Asynchronous return.

        """
        return self.data_to_server(amp.MsgPortal2Server, session.sessid, **kwargs)

    def send_AdminPortal2Server(self, session, operation="", **kwargs):
        """
        Send Admin instructions from the Portal to the Server.
        Executed on the Portal.

        Args:
            session (Session): Session.
            operation (char, optional): Identifier for the server operation, as defined by the
                global variables in `evennia/server/amp.py`.
            data (str or dict, optional): Data used in the administrative operation.

        """
        return self.data_to_server(
            amp.AdminPortal2Server, session.sessid, operation=operation, **kwargs
        )

    # receive amp data

    @amp.MsgStatus.responder
    @amp.catch_traceback
    def portal_receive_status(self, status):
        """
        Returns run-status for the server/portal.

        Args:
            status (str): Not used.
        Returns:
            status (dict): The status is a tuple
                (portal_running, server_running, portal_pid, server_pid).

        """
        # print('Received PSTATUS request')
        return {"status": amp.dumps(self.get_status())}

    @amp.MsgLauncher2Portal.responder
    @amp.catch_traceback
    def portal_receive_launcher2portal(self, operation, arguments):
        """
        Receives message arriving from evennia_launcher.
        This method is executed on the Portal.

        Args:
            operation (str): The action to perform.
            arguments (str): Possible argument to the instruction, or the empty string.

        Returns:
            result (dict): The result back to the launcher.

        Notes:
            This is the entrypoint for controlling the entire Evennia system from the evennia
            launcher. It can obviously only accessed when the Portal is already up and running.

        """
        # Since the launcher command uses amp.String() we need to convert from byte here.
        operation = str(operation, "utf-8")
        self.factory.launcher_connection = self
        _, server_connected, _, _, _, _ = self.get_status()

        # logger.log_msg("Evennia Launcher->Portal operation %s:%s received" % (ord(operation), arguments))

        # logger.log_msg("operation == amp.SSTART: {}: {}".format(operation == amp.SSTART, amp.loads(arguments)))

        if operation == amp.SSTART:  # portal start  #15
            # first, check if server is already running
            if not server_connected:
                self.wait_for_server_connect(self.send_Status2Launcher)
                self.start_server(amp.loads(arguments))

        elif operation == amp.SRELOAD:  # reload server #14
            if server_connected:
                # We let the launcher restart us once they get the signal
                self.factory.server_connection.wait_for_disconnect(self.send_Status2Launcher)
                self.stop_server(mode="reload")
            else:
                self.wait_for_server_connect(self.send_Status2Launcher)
                self.start_server(amp.loads(arguments))

        elif operation == amp.SRESET:  # reload server #19
            if server_connected:
                self.factory.server_connection.wait_for_disconnect(self.send_Status2Launcher)
                self.stop_server(mode="reset")
            else:
                self.wait_for_server_connect(self.send_Status2Launcher)
                self.start_server(amp.loads(arguments))

        elif operation == amp.SSHUTD:  # server-only shutdown #17
            if server_connected:
                self.factory.server_connection.wait_for_disconnect(self.send_Status2Launcher)
                self.stop_server(mode="shutdown")

        elif operation == amp.PSHUTD:  # portal + server shutdown  #16
            if server_connected:
                self.factory.server_connection.wait_for_disconnect(self.factory.portal.shutdown)
            else:
                self.factory.portal.shutdown()

        else:
            logger.log_err("Operation {} not recognized".format(operation))
            raise Exception("operation %(op)s not recognized." % {"op": operation})

        return {}

    @amp.MsgServer2Portal.responder
    @amp.catch_traceback
    def portal_receive_server2portal(self, packed_data):
        """
        Receives message arriving to Portal from Server.
        This method is executed on the Portal.

        Args:
            packed_data (str): Pickled data (sessid, kwargs) coming over the wire.

        """
        try:
            sessid, kwargs = self.data_in(packed_data)
            session = evennia.PORTAL_SESSION_HANDLER.get(sessid, None)
            if session:
                evennia.PORTAL_SESSION_HANDLER.data_out(session, **kwargs)
        except Exception:
            logger.log_trace("packed_data len {}".format(len(packed_data)))
        return {}

    @amp.AdminServer2Portal.responder
    @amp.catch_traceback
    def portal_receive_adminserver2portal(self, packed_data):
        """

        Receives and handles admin operations sent to the Portal
        This is executed on the Portal.

        Args:
            packed_data (str): Data received, a pickled tuple (sessid, kwargs).

        """
        self.factory.server_connection = self

        sessid, kwargs = self.data_in(packed_data)

        # logger.log_msg("Evennia Server->Portal admin data %s:%s received" % (sessid, kwargs))

        operation = kwargs.pop("operation")
        portal_sessionhandler = evennia.PORTAL_SESSION_HANDLER

        if operation == amp.SLOGIN:  # server_session_login
            # a session has authenticated; sync it.
            session = portal_sessionhandler.get(sessid)
            if session:
                portal_sessionhandler.server_logged_in(session, kwargs.get("sessiondata"))

        elif operation == amp.SDISCONN:  # server_session_disconnect
            # the server is ordering to disconnect the session
            session = portal_sessionhandler.get(sessid)
            if session:
                portal_sessionhandler.server_disconnect(session, reason=kwargs.get("reason"))

        elif operation == amp.SDISCONNALL:  # server_session_disconnect_all
            # server orders all sessions to disconnect
            portal_sessionhandler.server_disconnect_all(reason=kwargs.get("reason"))

        elif operation == amp.SRELOAD:  # server reload
            self.factory.server_connection.wait_for_disconnect(
                self.start_server, self.factory.portal.server_twistd_cmd
            )
            self.stop_server(mode="reload")

        elif operation == amp.SRESET:  # server reset
            self.factory.server_connection.wait_for_disconnect(
                self.start_server, self.factory.portal.server_twistd_cmd
            )
            self.stop_server(mode="reset")

        elif operation == amp.SSHUTD:  # server-only shutdown
            self.stop_server(mode="shutdown")

        elif operation == amp.PSHUTD:  # full server+server shutdown
            self.factory.server_connection.wait_for_disconnect(self.factory.portal.shutdown)
            self.stop_server(mode="shutdown")

        elif operation == amp.PSYNC:  # portal sync
            # Server has (re-)connected and wants the session data from portal
            self.factory.portal.server_info_dict = kwargs.get("info_dict", {})
            self.factory.portal.server_process_id = kwargs.get("spid", None)
            # this defaults to 'shutdown' or whatever value set in server_stop
            server_restart_mode = self.factory.portal.server_restart_mode

            sessdata = evennia.PORTAL_SESSION_HANDLER.get_all_sync_data()
            self.send_AdminPortal2Server(
                amp.DUMMYSESSION,
                amp.PSYNC,
                server_restart_mode=server_restart_mode,
                sessiondata=sessdata,
                portal_start_time=self.factory.portal.start_time,
            )
            evennia.PORTAL_SESSION_HANDLER.at_server_connection()

            if self.factory.server_connection:
                # this is an indication the server has successfully connected, so
                # we trigger any callbacks (usually to tell the launcher server is up)
                for callback, args, kwargs in self.factory.server_connect_callbacks:
                    try:
                        callback(*args, **kwargs)
                    except Exception:
                        logger.log_trace()
                self.factory.server_connect_callbacks = []

        elif operation == amp.SSYNC:  # server_session_sync
            # server wants to save session data to the portal,
            # maybe because it's about to shut down.
            portal_sessionhandler.server_session_sync(
                kwargs.get("sessiondata"), kwargs.get("clean", True)
            )

            # set a flag in case we are about to shut down soon
            self.factory.server_restart_mode = True

        elif operation == amp.SCONN:  # server_force_connection (for irc/etc)
            portal_sessionhandler.server_connect(**kwargs)

        else:
            raise Exception("operation %(op)s not recognized." % {"op": operation})
        return {}



# -*- coding: utf-8 -*-
# Generated by Django 1.9.11 on 2017-01-25 22:30


import re

from django.db import migrations


def update_perms_and_locks(apps, schema_editor):

    # update all permissions
    Tag = apps.get_model("typeclasses", "Tag")
    perm_map = {
        "guests": "guest",
        "players": "player",
        "playerhelpers": "helper",
        "builders": "builder",
        "wizards": "admin",
        "immortals": "developer",
    }

    for perm in Tag.objects.filter(db_tagtype="permission"):
        if perm.db_key in perm_map:
            perm.db_key = perm_map[perm.db_key]
            perm.save(update_fields=("db_key",))

    # update all locks on all entities
    apps_models = [
        ("objects", "ObjectDB"),
        ("accounts", "AccountDB"),
        ("scripts", "ScriptDB"),
        ("comms", "ChannelDB"),
    ]
    p_reg = re.compile(
        r"(?<=perm\()(\w+)(?=\))|(?<=perm_above\()(\w+)(?=\))", re.IGNORECASE + re.UNICODE
    )

    def _sub(match):
        perm = match.group(1)
        return perm_map[perm.lower()].capitalize() if (perm and perm.lower() in perm_map) else perm

    for app_tuple in apps_models:
        TClass = apps.get_model(*app_tuple)
        for obj in TClass.objects.filter(db_lock_storage__icontains="perm"):
            orig_lock = obj.db_lock_storage
            repl_lock = p_reg.sub(_sub, orig_lock)
            if repl_lock != orig_lock:
                obj.db_lock_storage = repl_lock
                obj.save(update_fields=("db_lock_storage",))


class Migration(migrations.Migration):

    dependencies = [("typeclasses", "0007_tag_migrations_may_be_slow")]

    operations = [migrations.RunPython(update_perms_and_locks)]



"""
This module implements the telnet protocol.

This depends on a generic session module that implements
the actual login procedure of the game, tracks
sessions etc.

"""

import re

from django.conf import settings
from twisted.conch.telnet import (
    ECHO,
    GA,
    IAC,
    LINEMODE,
    LINEMODE_EDIT,
    LINEMODE_TRAPSIG,
    MODE,
    NOP,
    NULL,
    WILL,
    WONT,
    StatefulTelnetProtocol,
    Telnet,
)
from twisted.internet import protocol
from twisted.internet.task import LoopingCall

from evennia.server.portal import mssp, naws, suppress_ga, telnet_oob, ttype
from evennia.server.portal.mccp import MCCP, Mccp, mccp_compress
from evennia.server.portal.mxp import Mxp, mxp_parse
from evennia.utils import ansi
from evennia.utils.utils import class_from_module, to_bytes

_RE_N = re.compile(r"\|n$")
_RE_LEND = re.compile(rb"\n$|\r$|\r\n$|\r\x00$|", re.MULTILINE)
_RE_LINEBREAK = re.compile(rb"\n\r|\r\n|\n|\r", re.DOTALL + re.MULTILINE)
_RE_SCREENREADER_REGEX = re.compile(
    r"%s" % settings.SCREENREADER_REGEX_STRIP, re.DOTALL + re.MULTILINE
)
_IDLE_COMMAND = str.encode(settings.IDLE_COMMAND + "\n")

# identify HTTP indata
_HTTP_REGEX = re.compile(
    b"(GET|HEAD|POST|PUT|DELETE|TRACE|OPTIONS|CONNECT|PATCH) (.*? HTTP/[0-9]\.[0-9])", re.I
)

_HTTP_WARNING = bytes(
    """
    This is Evennia's Telnet port and cannot be used for regular HTTP traffic.
    Use a telnet client to connect here and point your browser to the server's
    dedicated web port instead.

    """.strip(),
    "utf-8",
)


_BASE_SESSION_CLASS = class_from_module(settings.BASE_SESSION_CLASS)


class TelnetServerFactory(protocol.ServerFactory):
    """
    This exists only to name this better in logs.

    """

    noisy = False

    def logPrefix(self):
        return "Telnet"


class TelnetProtocol(Telnet, StatefulTelnetProtocol, _BASE_SESSION_CLASS):
    """
    Each player connecting over telnet (ie using most traditional mud
    clients) gets a telnet protocol instance assigned to them.  All
    communication between game and player goes through here.

    """

    def __init__(self, *args, **kwargs):
        self.protocol_key = "telnet"
        super().__init__(*args, **kwargs)

    def dataReceived(self, data):
        """
        Unused by default, but a good place to put debug printouts
        of incoming data.

        """
        # print(f"telnet dataReceived: {data}")
        try:
            super().dataReceived(data)
        except ValueError as err:
            from evennia.utils import logger

            logger.log_err(f"Malformed telnet input: {err}")

    def connectionMade(self):
        """
        This is called when the connection is first established.

        """
        # important in order to work normally with standard telnet
        self.do(LINEMODE).addErrback(self._wont_linemode)
        # initialize the session
        self.line_buffer = b""
        client_address = self.transport.client
        client_address = client_address[0] if client_address else None
        # this number is counted down for every handshake that completes.
        # when it reaches 0 the portal/server syncs their data
        self.handshakes = 8  # suppress-go-ahead, naws, ttype, mccp, mssp, msdp, gmcp, mxp

        self.init_session(self.protocol_key, client_address, self.factory.sessionhandler)
        self.protocol_flags["ENCODING"] = settings.ENCODINGS[0] if settings.ENCODINGS else "utf-8"
        # add this new connection to sessionhandler so
        # the Server becomes aware of it.
        self.sessionhandler.connect(self)
        # change encoding to ENCODINGS[0] which reflects Telnet default encoding

        # suppress go-ahead
        self.sga = suppress_ga.SuppressGA(self)
        # negotiate client size
        self.naws = naws.Naws(self)
        # negotiate ttype (client info)
        # Obs: mudlet ttype does not seem to work if we start mccp before ttype. /Griatch
        self.ttype = ttype.Ttype(self)
        # negotiate mccp (data compression) - turn this off for wireshark analysis
        self.mccp = Mccp(self)
        # negotiate mssp (crawler communication)
        self.mssp = mssp.Mssp(self)
        # oob communication (MSDP, GMCP) - two handshake calls!
        self.oob = telnet_oob.TelnetOOB(self)
        # mxp support
        self.mxp = Mxp(self)

        from evennia.utils.utils import delay

        # timeout the handshakes in case the client doesn't reply at all
        self._handshake_delay = delay(2, callback=self.handshake_done, timeout=True)

        # TCP/IP keepalive watches for dead links
        self.transport.setTcpKeepAlive(1)
        # The TCP/IP keepalive is not enough for some networks;
        # we have to complement it with a NOP keep-alive.
        self.protocol_flags["NOPKEEPALIVE"] = True
        self.nop_keep_alive = None
        self.toggle_nop_keepalive()

    def _wont_linemode(self, *args):
        """
        Client refuses do(linemode). This is common for MUD-specific
        clients, but we must ask for the sake of raw telnet. We ignore
        this error.

        """
        pass

    def _send_nop_keepalive(self):
        """
        Send NOP keepalive unless flag is set

        """
        if self.protocol_flags.get("NOPKEEPALIVE"):
            self._write(IAC + NOP)

    def toggle_nop_keepalive(self):
        """
        Allow to toggle the NOP keepalive for those sad clients that
        can't even handle a NOP instruction. This is turned off by the
        protocol_flag NOPKEEPALIVE (settable e.g. by the default
        `option` command).

        """
        if self.nop_keep_alive and self.nop_keep_alive.running:
            self.nop_keep_alive.stop()
        else:
            self.nop_keep_alive = LoopingCall(self._send_nop_keepalive)
            self.nop_keep_alive.start(30, now=False)

    def handshake_done(self, timeout=False):
        """
        This is called by all telnet extensions once they are finished.
        When all have reported, a sync with the server is performed.
        The system will force-call this sync after a small time to handle
        clients that don't reply to handshakes at all.

        """
        if timeout:
            if self.handshakes > 0:
                self.handshakes = 0
                self.sessionhandler.sync(self)
        else:
            self.handshakes -= 1
            if self.handshakes <= 0:
                # do the sync
                self.sessionhandler.sync(self)

    def at_login(self):
        """
        Called when this session gets authenticated by the server.

        """
        pass

    def enableRemote(self, option):
        """
        This sets up the remote-activated options we allow for this protocol.

        Args:
            option (char): The telnet option to enable.

        Returns:
            enable (bool): If this option should be enabled.

        """
        if option == LINEMODE:
            # make sure to activate line mode with local editing for all clients
            self.requestNegotiation(
                LINEMODE, MODE + bytes(chr(ord(LINEMODE_EDIT) + ord(LINEMODE_TRAPSIG)), "ascii")
            )
            return True
        else:
            return (
                option == ttype.TTYPE
                or option == naws.NAWS
                or option == MCCP
                or option == mssp.MSSP
                or option == ECHO
                or option == suppress_ga.SUPPRESS_GA
            )

    def disableRemote(self, option):
        return (
            option == LINEMODE
            or option == ttype.TTYPE
            or option == naws.NAWS
            or option == MCCP
            or option == mssp.MSSP
            or option == ECHO
            or option == suppress_ga.SUPPRESS_GA
        )

    def enableLocal(self, option):
        """
        Call to allow the activation of options for this protocol

        Args:
            option (char): The telnet option to enable locally.

        Returns:
            enable (bool): If this option should be enabled.

        """
        return (
            option == LINEMODE
            or option == MCCP
            or option == ECHO
            or option == suppress_ga.SUPPRESS_GA
        )

    def disableLocal(self, option):
        """
        Disable a given option locally.

        Args:
            option (char): The telnet option to disable locally.

        """
        if option == LINEMODE:
            return True
        if option == ECHO:
            return True
        if option == MCCP:
            self.mccp.no_mccp(option)
            return True
        else:
            try:
                return super().disableLocal(option)
            except Exception:
                from evennia.utils import logger

                logger.log_trace()

    def connectionLost(self, reason):
        """
        this is executed when the connection is lost for whatever
        reason. it can also be called directly, from the disconnect
        method

        Args:
            reason (str): Motivation for losing connection.

        """
        self.sessionhandler.disconnect(self)
        self.transport.loseConnection()

    def applicationDataReceived(self, data):
        """
        Telnet method called when non-telnet-command data is coming in
        over the telnet connection. We pass it on to the game engine
        directly.

        Args:
            data (str): Incoming data.

        """
        if not data:
            data = [data]
        elif data.strip() == NULL:
            # this is an ancient type of keepalive used by some
            # legacy clients. There should never be a reason to send a
            # lone NULL character so this seems to be a safe thing to
            # support for backwards compatibility. It also stops the
            # NULL from continuously popping up as an unknown command.
            data = [_IDLE_COMMAND]
        else:
            data = _RE_LINEBREAK.split(data)

            if len(data) > 2 and _HTTP_REGEX.match(data[0]):
                # guard against HTTP request on the Telnet port; we
                # block and kill the connection.
                self.transport.write(_HTTP_WARNING)
                self.transport.loseConnection()
                return

            if self.line_buffer and len(data) > 1:
                # buffer exists, it is terminated by the first line feed
                data[0] = self.line_buffer + data[0]
                self.line_buffer = b""
            # if the last data split is empty, it means all splits have
            # line breaks, if not, it is unterminated and must be
            # buffered.
            self.line_buffer += data.pop()
        # send all data chunks
        for dat in data:
            self.data_in(text=dat + b"\n")

    def _write(self, data):
        """
        Hook overloading the one used in plain telnet

        """
        data = data.replace(b"\n", b"\r\n").replace(b"\r\r\n", b"\r\n")
        super()._write(mccp_compress(self, data))

    def sendLine(self, line):
        """
        Hook overloading the one used by linereceiver.

        Args:
            line (str): Line to send.

        """
        line = to_bytes(line, self)
        # escape IAC in line mode, and correctly add \r\n (the TELNET end-of-line)
        line = line.replace(IAC, IAC + IAC)
        line = line.replace(b"\n", b"\r\n")
        if not line.endswith(b"\r\n") and self.protocol_flags.get("FORCEDENDLINE", True):
            line += b"\r\n"
        if not self.protocol_flags.get("NOGOAHEAD", True):
            line += IAC + GA
        return self.transport.write(mccp_compress(self, line))

    # Session hooks

    def disconnect(self, reason=""):
        """
        Generic hook for the engine to call in order to
        disconnect this protocol.

        Args:
            reason (str, optional): Reason for disconnecting.

        """
        self.data_out(text=((reason,), {}))
        self.connectionLost(reason)

    def data_in(self, **kwargs):
        """
        Data User -> Evennia

        Keyword Args:
            kwargs (any): Options from the protocol.

        """
        # from evennia.server.profiling.timetrace import timetrace  # DEBUG
        # text = timetrace(text, "telnet.data_in")  # DEBUG

        self.sessionhandler.data_in(self, **kwargs)

    def data_out(self, **kwargs):
        """
        Data Evennia -> User

        Keyword Args:
            kwargs (any): Options to the protocol

        """
        self.sessionhandler.data_out(self, **kwargs)

    # send_* methods

    def send_text(self, *args, **kwargs):
        """
        Send text data. This is an in-band telnet operation.

        Args:
            text (str): The first argument is always the text string to send. No other arguments
                are considered.
        Keyword Args:
            options (dict): Send-option flags

               - mxp: Enforce MXP link support.
               - ansi: Enforce no ANSI colors.
               - xterm256: Enforce xterm256 colors, regardless of TTYPE.
               - noxterm256: Enforce no xterm256 color support, regardless of TTYPE.
               - nocolor: Strip all Color, regardless of ansi/xterm256 setting.
               - raw: Pass string through without any ansi processing
                    (i.e. include Evennia ansi markers but do not
                    convert them into ansi tokens)
               - echo: Turn on/off line echo on the client. Turn
                    off line echo for client, for example for password.
                    Note that it must be actively turned back on again!

        """
        text = args[0] if args else ""
        if text is None:
            return

        # handle arguments
        options = kwargs.get("options", {})
        flags = self.protocol_flags
        xterm256 = options.get(
            "xterm256", flags.get("XTERM256", False) if flags.get("TTYPE", False) else True
        )
        useansi = options.get(
            "ansi", flags.get("ANSI", False) if flags.get("TTYPE", False) else True
        )
        raw = options.get("raw", flags.get("RAW", False))
        nocolor = options.get("nocolor", flags.get("NOCOLOR") or not (xterm256 or useansi))
        echo = options.get("echo", None)
        mxp = options.get("mxp", flags.get("MXP", False))
        screenreader = options.get("screenreader", flags.get("SCREENREADER", False))

        if screenreader:
            # screenreader mode cleans up output
            text = ansi.parse_ansi(text, strip_ansi=True, xterm256=False, mxp=False)
            text = _RE_SCREENREADER_REGEX.sub("", text)

        if options.get("send_prompt"):
            # send a prompt instead.
            prompt = text
            if not raw:
                # processing
                prompt = ansi.parse_ansi(
                    _RE_N.sub("", prompt) + ("||n" if prompt.endswith("|") else "|n"),
                    strip_ansi=nocolor,
                    xterm256=xterm256,
                )
                if mxp:
                    prompt = mxp_parse(prompt)
            prompt = to_bytes(prompt, self)
            prompt = prompt.replace(IAC, IAC + IAC).replace(b"\n", b"\r\n")
            if not self.protocol_flags.get(
                "NOPROMPTGOAHEAD", self.protocol_flags.get("NOGOAHEAD", True)
            ):
                prompt += IAC + GA
            self.transport.write(mccp_compress(self, prompt))
        else:
            if echo is not None:
                # turn on/off echo. Note that this is a bit turned around since we use
                # echo as if we are "turning off the client's echo" when telnet really
                # handles it the other way around.
                if echo:
                    # by telling the client that WE WON'T echo, the client knows
                    # that IT should echo. This is the expected behavior from
                    # our perspective.
                    self.transport.write(mccp_compress(self, IAC + WONT + ECHO))
                else:
                    # by telling the client that WE WILL echo, the client can
                    # safely turn OFF its OWN echo.
                    self.transport.write(mccp_compress(self, IAC + WILL + ECHO))
            if raw:
                # no processing
                self.sendLine(text)
                return
            else:
                # we need to make sure to kill the color at the end in order
                # to match the webclient output.
                linetosend = ansi.parse_ansi(
                    _RE_N.sub("", text) + ("||n" if text.endswith("|") else "|n"),
                    strip_ansi=nocolor,
                    xterm256=xterm256,
                    mxp=mxp,
                )
                if mxp:
                    linetosend = mxp_parse(linetosend)
                self.sendLine(linetosend)

    def send_prompt(self, *args, **kwargs):
        """
        Send a prompt - a text without a line end. See send_text for argument options.

        """
        kwargs["options"].update({"send_prompt": True})
        self.send_text(*args, **kwargs)

    def send_default(self, cmdname, *args, **kwargs):
        """
        Send other oob data

        """
        if not cmdname == "options":
            self.oob.data_out(cmdname, *args, **kwargs)
